"""sentry utilities."""

import logging
import os
import platform
import typing as ty
import warnings
from contextlib import suppress
from functools import lru_cache
from importlib import metadata
from subprocess import run

from loguru import logger
from sentry_sdk.integrations.loguru import LoggingLevels, LoguruIntegration

try:
    from rich import print as pprint
except ImportError:  # pragma: no cover
    from pprint import pprint

from getpass import getuser

import sentry_sdk
from koyo.utilities import running_as_pyinstaller_app

sentry_sdk.set_user({"username": getuser(), "ip_address": "{{auto}}"})

# disable logging from sentry_sdk
logging.getLogger("sentry_sdk").setLevel(logging.ERROR)
warnings.filterwarnings("ignore", category=ResourceWarning, module="sentry_sdk")

VERSION: str = os.getenv("QTEXTRA_TELEMETRY_VERSION", "")
SENTRY_DSN: str = os.getenv("QTEXTRA_TELEMETRY_SENTRY_DSN", "")
SHOW_HOSTNAME: bool = os.getenv("QTEXTRA_TELEMETRY_SHOW_HOSTNAME", "0") in ("1", "True")
SHOW_LOCALS: bool = os.getenv("QTEXTRA_TELEMETRY_SHOW_LOCALS", "1") in ("1", "True")
DEBUG: bool = os.getenv("QTEXTRA_TELEMETRY_DEBUG") in ("1", "True")
PACKAGE: str = os.getenv("QTEXTRA_TELEMETRY_PACKAGE", "qtextra")


def strip_sensitive_data(event: dict, hint: dict):
    """Pre-send hook to strip sensitive data from `event` dict.
    https://docs.sentry.io/platforms/python/configuration/filtering/#filtering-error-events.
    """
    # modify event here

    # strip `abs_paths` from stack_trace to hide local paths
    with suppress(KeyError, IndexError):
        for exc in event["exception"]["values"]:
            for frame in exc["stacktrace"]["frames"]:
                frame.pop("abs_path", None)
        # only include the name of the executable in sys.argv (remove paths)
        if args := event["extra"]["sys.argv"]:
            args[0] = args[0].split(os.sep)[-1]
    if DEBUG:  # pragma: no cover
        pprint(event)
    logger.debug("Sending sentry event")
    return event


def is_editable_install(dist_name: str) -> bool:
    """Return True if `dist_name` is installed as editable.
    i.e: if the package isn't in site-packages or user site-packages.
    """
    from site import getsitepackages, getusersitepackages

    dist = metadata.distribution(dist_name)
    installed_paths = [*getsitepackages(), getusersitepackages()]
    root = str(dist.locate_file(""))
    return all(loc not in root for loc in installed_paths)


def try_get_git_sha(package: ty.Optional[str] = None) -> str:
    """Try to return a git sha, for `dist_name` and detect if dirty.
    Return empty string on failure.
    """
    package = package or PACKAGE
    try:
        ff = metadata.distribution(package).locate_file("")
        out = run(["git", "-C", ff, "rev-parse", "HEAD"], capture_output=True)
        if out.returncode:  # pragma: no cover
            return ""
        sha = out.stdout.decode().strip()
        # exit with 1 if there are differences and 0 means no differences
        # disallow external diff drivers
        out = run(["git", "-C", ff, "diff", "--no-ext-diff", "--quiet", "--exit-code"])
        if out.returncode:  # pragma: no cover
            sha += "-dirty"
        return sha
    except Exception:  # pragma: no cover
        return ""


@lru_cache
def get_release(package: ty.Optional[str] = None) -> str:
    """Get the current release string for `package`.
    If the package is an editable install, it will return the current git sha.
    Otherwise return version string from package metadata.
    """
    package = package or PACKAGE
    try:
        with suppress(ModuleNotFoundError, ImportError):
            if is_editable_install(package):
                sha = try_get_git_sha(package)
                if sha:
                    return sha
            return metadata.version(package)
    except Exception:
        pass
    return "UNDETECTED"


def _get_tags() -> dict:
    tags = {"platform.system": platform.system()}

    with suppress(ImportError):
        from napari.utils.info import _sys_name

        sys = _sys_name()
        if sys:
            tags["system_name"] = sys

    with suppress(ImportError):
        import qtpy

        tags["qtpy.API_NAME"] = qtpy.API_NAME
        tags["qtpy.QT_VERSION"] = qtpy.QT_VERSION

    with suppress(ModuleNotFoundError, ImportError):
        tags["editable_install"] = str(is_editable_install(PACKAGE))
    tags["frozen"] = running_as_pyinstaller_app()
    tags["username"] = getuser()
    return tags


def get_sample_event(**kwargs) -> dict:
    """Return an example event as would be generated by an exception."""
    EVENT = {}

    def _transport(event: dict):
        nonlocal EVENT
        EVENT = event

    settings = SENTRY_SETTINGS.copy()
    settings["release"] = get_release()
    settings["dsn"] = ""
    settings["transport"] = _transport
    settings.update(kwargs)

    with sentry_sdk.Client(**settings) as client:
        with sentry_sdk.Hub(client) as hub:
            # remove locals that wouldn't really be there
            del settings, _transport, kwargs, client
            try:
                some_variable = 1  # noqa
                another_variable = "my_string"  # noqa
                1 / 0  # noqa
            except Exception:
                with sentry_sdk.new_scope() as scope:
                    for k, v in _get_tags().items():
                        scope.set_tag(k, v)
                    del v, k, scope
                    hub.capture_exception()

    with suppress(KeyError, IndexError):
        # remove the mock hub from the event
        frames = EVENT["exception"]["values"][0]["stacktrace"]["frames"]
        del frames[-1]["vars"]["hub"]
    return EVENT


SENTRY_LOGURU = LoguruIntegration(
    level=LoggingLevels.CRITICAL.value,  # Capture info and above as breadcrumbs
    event_level=LoggingLevels.CRITICAL.value,  # Send errors as events
)

INTEGRATIONS = [SENTRY_LOGURU]


SENTRY_SETTINGS = {
    "dsn": SENTRY_DSN,
    "release": VERSION,
    # When enabled, local variables are sent along with stackframes.
    # This can have a performance and PII impact.
    # Enabled by default on platforms where this is available.
    "include_local_variables": SHOW_LOCALS,
    # A number between 0 and 1, controlling the percentage chance
    # a given transaction will be sent to Sentry.
    # (0 represents 0% while 1 represents 100%.)
    # Applies equally to all transactions created in the app.
    # Either this or traces_sampler must be defined to enable tracing.
    "enable_tracing": True,
    "traces_sample_rate": 1.0,
    # When provided, the name of the server is sent along and persisted
    # in the event. For many integrations the server name actually
    # corresponds to the device hostname, even in situations where the
    # machine is not actually a server. Most SDKs will attempt to
    # auto-discover this value. (computer name: potentially PII)
    # "server_name": "",
    # If this flag is enabled, certain personally identifiable information (PII)
    # is added by active integrations. By default, no such data is sent.
    "send_default_pii": True,
    # This function is called with an SDK-specific event object, and can return a
    # modified event object or nothing to skip reporting the event.
    # This can be used, for instance, for manual PII stripping before sending.
    "before_send": strip_sensitive_data,
    "debug": DEBUG,
    # -------------------------
    "environment": platform.platform(),
    # max_breadcrumbs=DEFAULT_MAX_BREADCRUMBS,
    # shutdown_timeout=2,
    "integrations": INTEGRATIONS,
    # in_app_include=[],
    # in_app_exclude=[],
    # default_integrations=True,
    # dist=None,
    # transport=None,
    # transport_queue_size=DEFAULT_QUEUE_SIZE,
    # sample_rate=1.0,
    # http_proxy=None,
    # https_proxy=None,
    # ignore_errors=[],
    # max_request_body_size="medium",
    # before_breadcrumb=None,
    # attach_stacktrace=False,
    # ca_certs=None,
    # propagate_traces=True,
    # traces_sampler=None,
    # auto_session_tracking=True,
    "auto_enabling_integrations": False,
    "profiles_sample_rate": 1.0,
    "profiler_mode": "thread",
    # "_experiments": {
    #     "profiles_sample_rate": 1.0,
    # },
}



"""Common widget list interface."""

from __future__ import annotations

import typing as ty
from contextlib import contextmanager

from koyo.timer import MeasureTimer
from loguru import logger
from qtpy.QtCore import Signal, Slot  # type: ignore[attr-defined]
from qtpy.QtWidgets import QFrame, QListWidget, QListWidgetItem, QSizePolicy, QWidget

import qtextra.helpers as hp

_W = ty.TypeVar("_W")  # Widget
_M = ty.TypeVar("_M")  # Model


class QtListItem(QFrame):
    """List item that is shown inside the QtListWidget."""

    # event triggered whenever an item is checked
    _evt_checked = Signal(QListWidgetItem, bool)
    # event triggered whenever an item is removed
    evt_remove = Signal(QListWidgetItem)
    # event triggered when double click occurred
    evt_double_clicked = Signal(QListWidgetItem, bool)
    # event triggered whenever the item is active
    evt_active = Signal(QListWidgetItem)

    # Attributes
    item: QListWidgetItem = None
    _is_checked: bool = False
    _mode: bool = False

    def _set_from_model(self, _=None):
        """Update UI elements."""
        raise NotImplementedError("Must implement method")

    @property
    def item_model(self) -> _M:
        """Get item model."""
        return self.item.item_model

    @item_model.setter
    def item_model(self, item_model: _M):
        """Update item model."""
        self.item.item_model = item_model
        self._set_from_model()

    @property
    def is_checked(self) -> bool:
        """Get check state."""
        return self._is_checked

    @property
    def name(self):
        """Get heatmap name."""
        return self.name_label.text()

    @property
    def hash_id(self):
        """Get hash id information the selected item."""
        return self.item_model.name

    @property
    def mode(self):
        """Setup mode."""
        return self._mode

    @mode.setter
    def mode(self, value: bool):
        self._mode = value
        self.setProperty("mode", str(value))
        hp.polish_widget(self)
        self.evt_active.emit(self.item)

    def set_state(self, state: bool):
        """Check."""
        state = bool(state)
        self._is_checked = state
        if hasattr(self, "check_label"):
            self.check_label.setVisible(state)
        elif hasattr(self, "checkbox"):
            with hp.qt_signals_blocked(self.checkbox):
                self.checkbox.setChecked(state)
        self.mode = str(state)
        self._evt_checked.emit(self.item, self._is_checked)

    def mouseDoubleClickEvent(self, event):
        """Detect double-click event."""
        self.evt_double_clicked.emit(self.item, self._is_checked)

    def refresh(self):
        """Refresh values in the widget."""
        self._set_from_model()
        self.parent().update()


class QtListWidget(QListWidget):
    """List of notifications."""

    evt_updated = Signal(int)
    evt_added = Signal(object)
    evt_pre_remove = Signal(object)
    evt_remove = Signal(object)
    evt_cleared = Signal()

    _is_setup = False

    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)
        self.setSpacing(1)
        self.setMinimumHeight(12)
        self.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.MinimumExpanding)
        self.setUniformItemSizes(True)

    def closeEvent(self, event):
        """Close event."""
        self.teardown()
        return super().closeEvent(event)

    def teardown(self) -> None:
        """Teardown method."""

    def refresh_list(self):
        """Refresh list of items. This method should be re-implemented by subclasses."""

    def _get_menu(self):
        menu = hp.make_menu(self, "Actions")
        menu_remove = hp.make_menu_item(self, "Refresh list...", menu=menu, icon="refresh")
        menu_remove.triggered.connect(self.refresh_list)

    def _get_check_state(self, state: bool, attr: str = "is_checked"):
        items = []
        for widget in self.widget_iter():
            if hasattr(widget, attr) and getattr(widget, attr) == state:
                items.append((widget.hash_id, state))
        return items

    def _make_widget(self, item: QListWidgetItem):
        raise NotImplementedError("Must implement method")

    def _check_existing(self, item_model: _M) -> bool:
        """This method should be modified actually implement checking functionality."""
        return False

    @property
    def n_rows(self) -> int:
        """Return the current number of rows in the widget."""
        return self.count()

    def widget_iter(self) -> ty.Iterator[_W]:
        """Iterate through list of widgets."""
        for index in range(self.count()):
            yield self.itemWidget(self.item(index))

    def item_iter(self, indices: ty.Sequence[int] | None = None, reverse: bool = False) -> ty.Iterator[_W]:
        """Iterate through list of widgets."""
        if indices is None:
            indices = range(self.count())

        iterator = indices if not reverse else reversed(indices)
        for index in iterator:
            yield self.item(index)  # type: ignore[misc]

    def item_model_widget_iter(self) -> ty.Iterator[tuple[QListWidgetItem, _M, _W]]:
        """Iterate through list of widgets."""
        for index in range(self.count()):
            item = self.item(index)
            widget = self.itemWidget(item)
            yield item, item.item_model, widget  # type: ignore[misc,union-attr]

    def model_iter(self, indices: ty.Sequence[int] | None = None) -> ty.Iterator[_M]:
        """Iterate through list of ions."""
        if indices is None:
            indices = range(self.count())
        for index in indices:
            item = self.item(index)
            if item:
                yield item.item_model

    def get_all_checked(self, *, reverse: bool = False) -> list[int]:
        """Get list of checked items."""
        checked = []
        for index, widget in enumerate(self.widget_iter()):  # type: ignore[var-annotated]
            if widget.is_checked:
                checked.append(index)
        if reverse:
            return list(reversed(checked))
        return checked

    def get_all_unchecked(self) -> list[int]:
        """Get list of checked items."""
        checked = []
        for index, widget in enumerate(self.widget_iter()):  # type: ignore[var-annotated]
            if not widget.is_checked:
                checked.append(index)
        return checked

    def get_hash_ids(self, indices: ty.Iterator[int]) -> list[str]:
        """Get list of names."""
        hash_ids = []
        for item_id in indices:
            item = self.item(item_id)
            hash_ids.append(item.item_model.name)  # type: ignore[union-attr]
        return hash_ids

    def get_item_widget_for_index(self, index: int) -> tuple[QListWidgetItem, _W]:
        """Get widget for specified item."""
        item = self.item(index)
        return item, self.itemWidget(item)

    def get_index_for_hash_id(self, hash_id: str) -> int:
        """Get index of the item."""
        for index, widget in enumerate(self.widget_iter()):  # type: ignore[var-annotated]
            if widget.hash_id == hash_id:
                return index
        return -1

    def get_item_model_for_index(self, index: int) -> _M:
        """Get item's model."""
        item = self.item(index)
        return item.item_model

    def get_item_for_item_model(self, item_model: _M) -> ty.Optional[QListWidgetItem]:
        """Get item by its model."""
        for item, _item_model, _ in self.item_model_widget_iter():  # type: ignore[var-annotated]
            if _item_model is item_model or _item_model == item_model:
                return item
        return None

    def get_widget_for_hash_id(self, hash_id: str) -> _W:
        """Return item's widget."""
        index = self.get_index_for_hash_id(hash_id)
        if index == -1:
            return None
        return self.get_item_widget_for_index(index)[1]

    def get_widget_for_item_model(self, item_model: _M) -> ty.Optional[_W]:
        """Get widget by its model."""
        for _, _item_model, widget in self.item_model_widget_iter():  # type: ignore[var-annotated]
            if _item_model is item_model or _item_model == item_model:
                return widget
        return None

    def get_hash_id_for_index(self, index: int) -> str:
        """Get item's hash id."""
        item = self.get_item_model_for_index(index)
        return item.name

    @Slot(QListWidgetItem)
    @Slot(QListWidgetItem, bool)
    def remove_item(self, item: QListWidgetItem, force: bool = False):
        """Remove item from the list."""
        self.evt_pre_remove.emit(item)
        self.takeItem(self.indexFromItem(item).row())
        self.evt_remove.emit(item.item_model)
        super().removeItemWidget(item)
        self.evt_updated.emit(self.count())

    def remove_by_index(self, index: int, force: bool = False):
        """Remove item from the list based on row id."""
        item = self.item(index)
        self.remove_item(item, force)

    def remove_by_item_model(self, item_model: _M, force: bool = False, **kwargs):
        """Remove item from the list based on the item model."""
        item = self.get_item_for_item_model(item_model)
        if item:
            self.remove_item(item, force, **kwargs)

    def move_item(self, index: int, new_index: int, item_model: _M = None):
        """Move item from one index to another."""
        item = self.takeItem(index)
        self.insertItem(new_index, item)

    def select_by_index(self, index: int):
        """Select item."""
        self.setCurrentIndex(index)

    def select_by_item_model(self, item_model):
        """Find by item model."""
        item = self.get_item_for_item_model(item_model)
        if item:
            self.select_by_item(item)

    def select_by_item(self, item: QListWidgetItem):
        """Select item."""
        self.setCurrentIndex(self.indexFromItem(item))

    def refresh(self) -> None:
        """Refresh widget UI."""
        for index in range(self.count()):
            widget = self.itemWidget(self.item(index))
            widget.refresh()

    def reset_data(self) -> None:
        """Reset data."""
        self.clear()
        self.evt_cleared.emit()

    def append_item(self, item_model: _M) -> tuple[ty.Optional[QListWidgetItem], ty.Optional[QWidget]]:
        """Append item."""
        if self._check_existing(item_model):
            return None, None
        try:
            item = QListWidgetItem(parent=self)
        except AttributeError:
            item = QListWidgetItem()
        item.item_model = item_model
        widget = self._make_widget(item)
        widget.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.MinimumExpanding)
        item.setSizeHint(widget.sizeHint())
        self.addItem(item)
        self.setItemWidget(item, widget)
        self.evt_added.emit(item_model)
        self.evt_updated.emit(self.count())
        return item, widget

    def insert_item(self, item_model: _M, index: int = 0) -> tuple[ty.Optional[QListWidgetItem], ty.Optional[QWidget]]:
        """Insert item in the list."""
        if self._check_existing(item_model):
            return None, None
        item = QListWidgetItem(parent=self)
        item.item_model = item_model
        self.insertItem(index, item)
        widget = self._make_widget(item)
        item.setSizeHint(widget.sizeHint())
        self.setItemWidget(item, widget)
        self.evt_updated.emit(self.count())
        return item, widget

    def _clear(self, _):
        self.clear()
        self.evt_updated.emit(self.count())

    @contextmanager
    def disable_updates(self):
        """Temporarily disable updates."""
        self.setUpdatesEnabled(False)
        yield
        self.setUpdatesEnabled(True)

    @contextmanager
    def measure_time(self, message: str = "Task took", print_: bool = False):
        """Measure time."""
        with MeasureTimer() as timer:
            yield
        msg = f"{message} {timer()}"
        logger.debug(msg)
        if print_:
            print(msg)



"""Toast."""

from __future__ import annotations

from contextlib import suppress

from qtpy.QtCore import Qt, QTimer
from qtpy.QtGui import QMouseEvent
from qtpy.QtWidgets import QProgressBar, QWidget

import qtextra.helpers as hp
from qtextra.widgets.qt_dialog import SubWindowBase
from qtextra.widgets.qt_label_icon import QtSeverityLabel


class QtToast(SubWindowBase):
    """Small popup notification that can contain actions."""

    # Animation attributes
    POSITION = "top_right"
    DISMISS_AFTER = 5000
    MAX_OPACITY = 1.0

    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent)

        self.timer_dismiss = QTimer()
        self.timer_remaining = QTimer()

        self.make_ui()
        if hasattr(parent, "evt_resized"):
            parent.evt_resized.connect(lambda: self.move_to(self.POSITION))

    # noinspection PyAttributeOutsideInit
    def make_ui(self) -> None:
        """Setup UI."""
        title_widget = QWidget()
        title_widget.setObjectName("toast_header")

        self._icon_label = QtSeverityLabel(title_widget)
        self._icon_label.set_xsmall()

        self._title_label = hp.make_label(title_widget, "", bold=True, object_name="transparent")
        hp.set_expanding_sizer_policy(self._title_label, True, False)

        self._date_label = hp.make_label(title_widget, "", object_name="transparent")

        self._close_btn = hp.make_qta_btn(title_widget, "cross", func=self.close)
        self._close_btn.set_xsmall()

        self._message_label = hp.make_label(self, "", wrap=True, enable_url=True)

        title_layout = hp.make_h_layout(parent=title_widget, margin=2, spacing=1)
        title_layout.addWidget(self._icon_label, alignment=Qt.AlignmentFlag.AlignVCenter)
        title_layout.addWidget(self._title_label, stretch=True, alignment=Qt.AlignmentFlag.AlignVCenter)
        title_layout.addWidget(self._date_label, alignment=Qt.AlignmentFlag.AlignVCenter)
        title_layout.addStretch(1)
        title_layout.addWidget(self._close_btn, alignment=Qt.AlignmentFlag.AlignTop)

        self._timer_indicator = QProgressBar(self)
        self._timer_indicator.setObjectName("progress_timer")
        self._timer_indicator.setTextVisible(False)

        # layout
        layout = hp.make_v_layout(parent=self, margin=0, spacing=0)
        layout.addWidget(title_widget)
        layout.addWidget(self._message_label, stretch=True)
        layout.addStretch(1)
        layout.addWidget(self._timer_indicator)

    def show_message(
        self, title: str, message: str, icon: str = "info", position: str = "top_right", duration: int = 5000
    ) -> None:
        """Show message."""
        self.POSITION = position
        self.DISMISS_AFTER = duration
        self._title_label.setText(title)
        self._title_label.adjustSize()
        self._message_label.setText(message)
        self._message_label.adjustSize()
        self._icon_label.severity = str(icon)
        self.adjustSize()
        self.move_to(self.POSITION)
        self.show()

    def show(self) -> None:
        """Show the message with a fade and slight slide in from the bottom."""

        def _update_timer_indicator() -> None:
            with suppress(RuntimeError):
                self._timer_indicator.setValue(int(self.timer_dismiss.remainingTime() / self.DISMISS_AFTER * 100))

        super().show()
        self.slide_in()
        if self.DISMISS_AFTER > 0:
            self.timer_dismiss.setInterval(self.DISMISS_AFTER)
            self.timer_dismiss.setSingleShot(True)
            self.timer_dismiss.timeout.connect(self.close)
            self.timer_dismiss.timeout.connect(self.timer_remaining.stop)
            self.timer_dismiss.start()

            self.timer_remaining.setInterval(50)
            self.timer_remaining.setSingleShot(False)
            self.timer_remaining.timeout.connect(_update_timer_indicator)
            self.timer_remaining.start()

    def mouseMoveEvent(self, event: QMouseEvent | None) -> None:  # type: ignore[override]
        """On hover, stop the self-destruct timer."""
        self.timer_dismiss.stop()
        self.timer_remaining.stop()
        self._timer_indicator.setVisible(False)
        return super().mouseMoveEvent(event)

    def deleteLater(self) -> None:
        """Stop all animations and timers before deleting."""
        self.timer_dismiss.stop()
        self.timer_remaining.stop()
        super().deleteLater()

    def close(self):
        self.timer_dismiss.stop()
        self.timer_remaining.stop()
        SubWindowBase.close(self)


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys
        from random import choice

        from qtextra.config import THEMES
        from qtextra.utils.dev import qframe

        def _popup_notif() -> None:
            pop = QtToast(frame)
            THEMES.set_theme_stylesheet(pop)
            # pop.show_message("Title", "Here is a message..")
            pop.show_message("Title", "Here is a message.\nA couple of lines long.\nAnother line")

        def _popup_notif3() -> None:
            from qtextra.helpers import toast_alt

            toast_alt(
                frame,
                "Title",
                "Here is a message.\nA couple of lines long.\nAnother line",
                icon=choice(["info", "warning", "error", "success"]),
            )

        def _popup_notif2() -> None:
            pop = QtToast(frame)
            THEMES.set_theme_stylesheet(pop)
            pop.show_message(
                "Title",
                (
                    "You can easily move images around by clicking inside the image and moving it left-right and"
                    "up-down.\nRotation is currently disabled and changes to scale and shearing will not be supported."
                ),
                duration=10000,
            )

        app, frame, ha = qframe(False, set_style=True)
        frame.setMinimumSize(600, 600)

        btn2 = hp.make_btn(frame, "Create random notification")
        btn2.clicked.connect(_popup_notif)
        ha.addWidget(btn2)
        btn2 = hp.make_btn(frame, "Create long notification")
        btn2.clicked.connect(_popup_notif2)
        ha.addWidget(btn2)
        btn2 = hp.make_btn(frame, "Create alternate notification")
        btn2.clicked.connect(_popup_notif3)
        ha.addWidget(btn2)
        ha.addWidget(btn2)
        ha.addStretch(1)

        frame.show()
        sys.exit(app.exec_())

    _main()



"""Confirm action by typing requested text."""

import typing as ty

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QDialog, QWidget

import qtextra.helpers as hp


class QtConfirmWithTextDialog(QDialog):
    """Confirm action by typing requested text."""

    def __init__(
        self,
        parent: ty.Optional[QWidget] = None,
        title: str = "Please confirm...",
        message: str = "Please type 'confirm' to continue.",
        request: str = "confirm",
    ):
        super().__init__(parent)
        self.setWindowFlags(self.windowFlags() | Qt.WindowStaysOnTopHint)  # type: ignore
        self.setObjectName("confirm_dialog")
        self.setMinimumSize(350, 200)
        self.setWindowTitle(title)
        self.request = request

        layout = hp.make_v_layout()
        layout.addWidget(hp.make_label(self, message, enable_url=True, wrap=True), stretch=True)
        self.request_edit = hp.make_line_edit(self, "", placeholder=request, func_changed=self.validate)
        layout.addWidget(self.request_edit, stretch=True)

        self.ok_btn = hp.make_btn(self, "Yes", func=self.accept)
        layout.addLayout(
            hp.make_h_layout(
                self.ok_btn,
                hp.make_btn(self, "No", func=self.reject, tooltip="Cancel and return to the app."),
            )
        )
        self.setLayout(layout)
        self.validate()
        self.request_edit.setFocus()

    def validate(self) -> None:
        """Validate the input."""
        enabled = self.request_edit.text() == self.request
        self.ok_btn.setEnabled(enabled)
        hp.update_widget_style(self.ok_btn, "success_btn" if enabled else "")



"""Splash screen."""

from koyo.typing import PathLike
from qtpy.QtCore import Qt, Slot  # type: ignore[attr-defined]
from qtpy.QtGui import QPixmap
from qtpy.QtWidgets import QSplashScreen

import qtextra.helpers as hp


class QtSplashScreen(QSplashScreen):
    """Splash screen."""

    TITLE = "qtextra"

    def __init__(self, path: PathLike, width: int = 360):
        from qtextra.config import EVENTS
        from qtextra.event_loop import get_app

        get_app()
        pm = QPixmap(path).scaled(
            width, width, Qt.AspectRatioMode.KeepAspectRatio, Qt.TransformationMode.SmoothTransformation
        )
        super().__init__(pm)
        self.show()
        hp.set_font(self, font_size=16)
        self.showMessage(f"Loading {self.TITLE}...", alignment=Qt.AlignmentFlag.AlignLeft, color=Qt.GlobalColor.black)

        EVENTS.evt_splash_msg.connect(self.on_message)
        EVENTS.evt_splash_close.connect(self.close)

    @Slot(str)  # type: ignore[misc]
    def on_message(self, msg: str):
        """Show message."""
        self.showMessage(msg, alignment=Qt.AlignmentFlag.AlignLeft, color=Qt.GlobalColor.white)


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.assets import ICONS
    from qtextra.utils.dev import qapplication

    app = qapplication()

    random_icon = next(iter(ICONS.values()))

    wdg = QtSplashScreen(random_icon)
    wdg.show()
    wdg.raise_()
    hp.call_later(wdg, wdg.close, delay=2000)
    sys.exit(app.exec_())



"""Checkable combobox."""

import typing as ty
from functools import partial

from qtpy.QtCore import QModelIndex, Qt, Signal
from qtpy.QtGui import QStandardItemModel
from qtpy.QtWidgets import QComboBox

from qtextra.helpers import call_later, qt_signals_blocked


class CheckableAbstractModel(QStandardItemModel):
    """Abstract model."""

    evt_checked = Signal(int, bool)

    def setData(self, index: QModelIndex, value: ty.Any, role: int = ...) -> bool:
        """Set data."""
        if role == Qt.ItemDataRole.CheckStateRole:
            # self.evt_checked.emit(index.row(), not value)
            call_later(self, partial(self.evt_checked.emit, index.row(), not value), delay=50)
        elif role == Qt.ItemDataRole.DisplayRole:
            pass
        return super().setData(index, value, role)


class QtCheckableComboBox(QComboBox):
    """Checkable combobox."""

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)
        self.setModel(CheckableAbstractModel(self))
        self.evt_checked = self.model().evt_checked

    def addItem(self, item: str, data: ty.Any = None):
        """Add item."""
        super().addItem(item, data)
        item = self.model().item(self.count() - 1, 0)
        item.setFlags(Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsEnabled)
        item.setCheckState(Qt.CheckState.Unchecked)

    def addItems(self, texts: ty.Sequence[str]) -> None:
        """Add items."""
        current = self.count()
        super().addItems(texts)
        for index in range(current, self.count()):
            item = self.model().item(index, 0)
            item.setFlags(Qt.ItemFlag.ItemIsUserCheckable | Qt.ItemFlag.ItemIsEnabled)
            item.setCheckState(Qt.CheckState.Unchecked)

    def itemChecked(self, index):
        """Item is checked."""
        item = self.model().item(index, 0)
        return item.checkState() == Qt.CheckState.Checked

    def setItemChecked(self, index: int, checked: bool) -> bool:
        """Set item checked."""
        item = self.model().item(index, 0)
        item.setCheckState(Qt.CheckState.Checked if checked else Qt.CheckState.Unchecked)

    def get_checked(self) -> ty.List[int]:
        """Get all checked items."""
        return [index for index in range(self.count()) if self.itemChecked(index)]

    def checked_texts(self) -> ty.List[str]:
        """Get text for all checked items."""
        return [self.itemText(index) for index in self.get_checked()]

    def checked_data(self):
        """Get data for all checked items."""
        return [self.itemData(index) for index in self.get_checked()]

    def set_checked_texts(self, texts: ty.List[str]):
        """Set checked texts."""
        with qt_signals_blocked(self):
            [self.setItemChecked(index, False) for index in range(self.count())]
            for text in texts:
                index = self.findText(text)
                if index != -1:
                    self.setItemChecked(index, True)

    # Alias methods to offer Qt-like interface
    checkedData = checked_data
    checkedTexts = checked_texts
    getChecked = get_checked
    setCheckedTexts = set_checked_texts


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    app, frame, ha = qframe(False)
    frame.setLayout(ha)
    frame.setMinimumSize(400, 400)

    wdg = QtCheckableComboBox()
    wdg.addItems(["Option 1", "Option 2", "Option 3"])
    ha.addWidget(wdg)

    frame.show()
    sys.exit(app.exec_())



"""Collapsible widget.

Taken from:
https://github.dev/napari/superqt/blob/f4d9881b0c64c0419fa2da182a1c403a01bd084f/src/superqt/collapsible/_collapsible.py
"""

from __future__ import annotations

from contextlib import suppress

from qtpy.QtCore import Qt
from qtpy.QtWidgets import QCheckBox, QHBoxLayout, QLayout, QWidget
from superqt import QCollapsible

import qtextra.helpers as hp
from qtextra.config import THEMES


class QtCheckCollapsible(QCollapsible):
    """A collapsible widget to hide and unhide child widgets.

    Based on https://stackoverflow.com/a/68141638
    """

    def __init__(
        self, title: str = "", parent: QWidget | None = None, icon: str = "info", warning_icon: str = "warning"
    ):
        super().__init__(title, parent)
        self.checkbox = QCheckBox()
        self.checkbox.stateChanged.connect(self._toggle_btn.setChecked)

        # remove button item from the layout
        self.layout().takeAt(0)

        self.action_btn = hp.make_qta_btn(self, icon_name=icon, standout=True, average=True)

        self.warning_label = hp.make_warning_label(self, "", icon_name=warning_icon, normal=True)

        # create layout where the first item is checkbox
        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(1)
        layout.addWidget(self.checkbox)
        layout.addWidget(self._toggle_btn, stretch=True)
        layout.addWidget(self.action_btn, alignment=Qt.AlignmentFlag.AlignVCenter)
        layout.addWidget(self.warning_label, alignment=Qt.AlignmentFlag.AlignVCenter)

        # add widget to layout
        self.layout().addLayout(layout)
        self.layout().setSpacing(0)
        self.layout().setContentsMargins(0, 0, 0, 0)
        self._update_icon()

        content = QWidget()
        layout = hp.make_form_layout(parent=content)
        layout.setContentsMargins(2, 2, 2, 2)
        self.setContent(content)
        with suppress(RuntimeError):
            THEMES.evt_theme_icon_changed.connect(self._update_icon)

    def _update_icon(self) -> None:
        self.setExpandedIcon(hp.make_qta_icon("chevron_down"))
        self.setCollapsedIcon(hp.make_qta_icon("chevron_up"))

    def set_checkbox_visible(self, state: bool) -> None:
        """Show or hide the checkbox."""
        self.checkbox.setVisible(state)

    def set_icon_visible(self, state: bool) -> None:
        """Show or hide the checkbox."""
        self.action_btn.setVisible(state)

    def set_warning_visible(self, state: bool) -> None:
        """Show or hide the checkbox."""
        self.warning_label.setVisible(state)

    @property
    def is_checked(self) -> bool:
        """Determine whether widget is checked."""
        return self.checkbox.isChecked()

    def _toggle(self):
        self.checkbox.setChecked(self._toggle_btn.isChecked())
        super()._toggle()

    def _checked(self):
        self._toggle_btn.setChecked(self.checkbox.isChecked())

    def addLayout(self, layout: QLayout):
        """Add layout to the central content widget's layout."""
        self._content.layout().addLayout(layout)

    def addRow(self, label: QWidget | QLayout, widget: QWidget | None = None):
        """Add layout to the central content widget's layout."""
        if not hasattr(self._content.layout(), "addRow"):
            raise ValueError("Layout does not have `addRow` method.")
        if widget:
            self._content.layout().addRow(label, widget)
        else:
            self._content.layout().addRow(label)

    # Alias methods to offer Qt-like interface
    setCheckboxVisible = set_checkbox_visible
    setIconVisible = set_icon_visible
    setWarningVisible = set_warning_visible


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys

        from qtextra.utils.dev import qframe

        app, frame, ha = qframe(False)
        frame.setMinimumSize(600, 600)

        wdg = QtCheckCollapsible(parent=frame)
        wdg.setText("Advanced options")
        ha.addWidget(wdg)

        wdg = QtCheckCollapsible(parent=frame)
        wdg.set_checkbox_visible(False)
        wdg.setText("Advanced options")
        ha.addWidget(wdg)

        frame.show()
        sys.exit(app.exec_())

    _main()



"""Close dialog with option to not ask again."""

from __future__ import annotations

import typing as ty

from koyo.system import IS_PYINSTALLER, IS_WIN
from qtpy.QtGui import QKeySequence
from qtpy.QtWidgets import QDialog, QHBoxLayout, QVBoxLayout, QWidget

import qtextra.helpers as hp


class QtConfirmCloseDialog(QDialog):
    """Confirm close dialog with an option to not ask again."""

    def __init__(
        self,
        parent: QWidget,
        attr: str | None = None,
        save_func: ty.Callable | None = None,
        config: ty.Optional[object] = None,
        no_icon: bool = False,
    ) -> None:
        super().__init__(parent)
        self.attr = attr
        self.config = config
        self.save_func = save_func

        if (IS_PYINSTALLER and IS_WIN) or no_icon:
            cancel_btn = hp.make_btn(self, "Cancel", tooltip="Cancel and return to the app.")
            save_btn = hp.make_btn(self, "Save", tooltip="Save and close the app.")
            close_btn = hp.make_btn(self, "Close", tooltip="Close the app.", properties={"type": "error"})
        else:
            cancel_btn = hp.make_qta_btn(
                self, "cancel", label="Cancel", standout=True, tooltip="Cancel and return to the app."
            )
            save_btn = hp.make_qta_btn(self, "save", label="Save", standout=True, tooltip="Save and close the app.")
            close_btn = hp.make_qta_btn(
                self, "warning", color="orange", label="Close", standout=True, tooltip="Close the app."
            )

        icon_label = hp.make_qta_label(self, "warning", color="orange")
        icon_label.set_xxxlarge()

        self.do_not_ask = hp.make_checkbox(self, "Do not ask in future")

        self.setWindowTitle("Close Application?")
        shortcut = QKeySequence("Ctrl+Q").toString(QKeySequence.SequenceFormat.NativeText)
        text = (
            f"Do you want to close the application? There might be some <b>unsaved</b> changes."
            f"<br><br>(<b>{shortcut}</b> to confirm)."
        )
        close_btn.setShortcut(QKeySequence("Ctrl+Q"))

        if callable(save_func):
            save_btn.clicked.connect(self.save_func)
        else:
            save_btn.hide()
        cancel_btn.clicked.connect(self.reject)
        close_btn.clicked.connect(self.accept)

        btn_layout = QHBoxLayout()
        btn_layout.addWidget(close_btn)
        btn_layout.addStretch(1)
        btn_layout.addWidget(save_btn)
        btn_layout.addWidget(cancel_btn)

        body_layout = QVBoxLayout()
        body_layout.addWidget(hp.make_label(self, text, enable_url=True, wrap=True))
        body_layout.addWidget(self.do_not_ask)
        body_layout.addLayout(btn_layout)

        icon_layout = QHBoxLayout(self)
        icon_layout.addWidget(icon_label)
        icon_layout.addLayout(body_layout)

        # for test purposes because of the problem with shortcut testing:
        # https://github.com/pytest-dev/pytest-qt/issues/254
        self.close_btn = close_btn
        self.cancel_btn = cancel_btn

    def accept(self):
        """Accept."""
        if self.do_not_ask.isChecked():
            if self.config and self.attr is not None:
                setattr(self.config, self.attr, False)
        return super().accept()


if __name__ == "__main__":  # pragma: no cover
    from qtextra.utils.dev import apply_style, qapplication

    app = qapplication()
    dlg = QtConfirmCloseDialog(None, no_icon=True)
    apply_style(dlg)
    dlg.exec_()



"""Typing utilities."""

import typing as ty
from enum import Enum

Callback = ty.Union[ty.Callable, ty.Sequence[ty.Callable]]
Orientation = ty.Literal["horizontal", "vertical"]
IconType = ty.Union[str, tuple[str, dict[str, ty.Any]]]


class TaskState(str, Enum):
    """State of the task."""

    QUEUED = "queued"
    RUN_NEXT = "run_next"
    RUNNING = "running"
    PAUSING = "pausing"
    PAUSED = "paused"
    INCOMPLETE = "incomplete"
    FINISHED = "finished"
    PART_FAILED = "part-failed"
    FAILED = "failed"
    CANCELLING = "cancelling"
    CANCELLED = "cancelled"
    LOCKED = "locked"


class WorkerState(str, Enum):
    """Worker state."""

    FINISHED = "finished"
    NOT_ENOUGH_SPACE = "not_enough_space"



from unittest.mock import patch

import pytest
from qtpy.QtCore import Qt

from qtextra.widgets.qt_button_icon import QtImagePushButton


@pytest.fixture
def setup_image_widget(qtbot):
    """Setup panel"""

    def _widget() -> QtImagePushButton:
        widget = QtImagePushButton()
        qtbot.addWidget(widget)
        return widget

    return _widget


class TestQtImagePushButton:
    right_click = 0

    def _on_right_click(self):
        self.right_click += 1

    def test_init(self, qtbot, setup_image_widget):
        widget = setup_image_widget()
        widget.setObjectName("info")
        assert widget

        with patch.object(widget, "on_click") as mock_click:
            qtbot.mouseClick(widget, Qt.LeftButton)
            mock_click.assert_called_once()

        with patch.object(widget, "on_right_click") as mock_click:
            qtbot.mouseClick(widget, Qt.RightButton)
            mock_click.assert_called_once()

        widget.connect_to_right_click(self._on_right_click)
        qtbot.mouseClick(widget, Qt.RightButton)
        assert self.right_click == 1



"""Button that copies contents of QTextEdit to the clipboard."""

from qtpy.QtGui import QGuiApplication, QImage
from qtpy.QtWidgets import QTextEdit

from qtextra.config import EVENTS, THEMES
from qtextra.widgets.qt_button_icon import QtImagePushButton


def copy_text_to_clipboard(text: str):
    """Helper function to easily copy text to clipboard while notifying the user."""
    cb = QGuiApplication.clipboard()
    cb.setText(text)
    EVENTS.evt_msg_success.emit("Copied text to clipboard!")


def copy_image_to_clipboard(image: QImage):
    """Helper function to easily copy image to clipboard while notifying the user."""
    cb = QGuiApplication.clipboard()
    cb.setImage(image)
    EVENTS.evt_msg_success.emit("Copied image to clipboard!")


class QtCopyToClipboardButton(QtImagePushButton):
    """Button to copy text box information to the clipboard.

    Parameters
    ----------
    text_edit : qtpy.QtWidgets.QTextEdit
        The text box contents linked to copy to clipboard button.

    Attributes
    ----------
    text_edit : qtpy.QtWidgets.QTextEdit
        The text box contents linked to copy to clipboard button.
    """

    def __init__(self, text_edit: QTextEdit):
        super().__init__()
        self.setObjectName("QtCopyToClipboardButton")
        self.text_edit = text_edit
        self.setToolTip("Copy to clipboard")
        self.set_qta("copy_to_clipboard")
        self.clicked.connect(self.copy_to_clipboard)

    def copy_to_clipboard(self) -> None:
        """Copy text to the clipboard."""
        from qtextra.helpers import add_flash_animation

        copy_text_to_clipboard(str(self.text_edit.toPlainText()))
        add_flash_animation(self.text_edit, color=THEMES.get_hex_color("foreground"), duration=500)

    # Alias methods to offer Qt-like interface
    copyToClipboard = copy_to_clipboard



"""
set default asyncio policy to be compatible with tornado.

Tornado 6 (at least) is not compatible with the default
asyncio implementation on Windows

Pick the older SelectorEventLoopPolicy on Windows
if the known-incompatible default policy is in use.

FIXME: if/when tornado supports the defaults in asyncio,
remove and bump tornado requirement for py38
borrowed from ipykernel:  https://github.com/ipython/ipykernel/pull/456
"""

import re
import sys
import typing as ty

from ipykernel.inprocess.ipkernel import InProcessInteractiveShell
from ipykernel.zmqshell import ZMQInteractiveShell
from IPython import get_ipython
from IPython.terminal.interactiveshell import TerminalInteractiveShell
from qtconsole.inprocess import QtInProcessKernelManager
from qtconsole.rich_jupyter_widget import RichJupyterWidget
from qtpy.QtCore import Qt
from qtpy.QtGui import QColor
from qtpy.QtWidgets import QFormLayout, QVBoxLayout, QWidget

from qtextra.widgets.qt_dialog import QtFramelessTool


def str_to_rgb(arg):
    """Convert an rgb string 'rgb(x,y,z)' to a list of ints [x,y,z]."""
    return list(map(int, re.match(r"rgb\((\d+),\s*(\d+),\s*(\d+)\)", arg).groups()))


if sys.platform.startswith("win"):
    import asyncio

    try:
        from asyncio import WindowsProactorEventLoopPolicy, WindowsSelectorEventLoopPolicy
    except ImportError:
        pass
        # not affected
    else:
        if type(asyncio.get_event_loop_policy()) is WindowsProactorEventLoopPolicy:
            # WindowsProactorEventLoopPolicy is not compatible with tornado 6
            # fallback to the pre-3.8 default of Selector
            asyncio.set_event_loop_policy(WindowsSelectorEventLoopPolicy())


class QtConsole(RichJupyterWidget):
    """Qt view for the console, an integrated iPython terminal in napari.

    Parameters
    ----------
    variables : dict, optional
        Dictionary of variables to add to the console namespace.

    Attributes
    ----------
    kernel_client : qtconsole.inprocess.QtInProcessKernelClient,
                    qtconsole.client.QtKernelClient, or None
        Client for the kernel if it exists, None otherwise.
    shell : ipykernel.inprocess.ipkernel.InProcessInteractiveShell,
            ipykernel.zmqshell.ZMQInteractiveShell, or None.
        Shell for the kernel if it exists, None otherwise.
    """

    def __init__(self, variables: ty.Optional[ty.Dict[str, ty.Any]] = None):
        super().__init__()
        # Connect theme update
        if variables is None:
            variables = {}

        # this makes calling `setFocus()` on a QtConsole give keyboard focus to
        # the underlying `QTextEdit` widget
        self.setFocusProxy(self._control)

        # get current running instance or create new instance
        shell = get_ipython()

        if shell is None:
            # If there is no currently running instance create an in-process
            # kernel.
            kernel_manager = QtInProcessKernelManager()
            kernel_manager.start_kernel(show_banner=False)
            kernel_manager.kernel.gui = "qt"

            kernel_client = kernel_manager.client()
            kernel_client.start_channels()

            self.kernel_manager = kernel_manager
            self.kernel_client = kernel_client
            self.shell = kernel_manager.kernel.shell
            self.push = self.shell.push
        elif type(shell) == InProcessInteractiveShell:
            # If there is an existing running InProcessInteractiveShell
            # it is likely because multiple viewers have been launched from
            # the same process. In that case create a new kernel.
            # Connect existing kernel
            kernel_manager = QtInProcessKernelManager(kernel=shell.kernel)
            kernel_client = kernel_manager.client()

            self.kernel_manager = kernel_manager
            self.kernel_client = kernel_client
            self.shell = kernel_manager.kernel.shell
            self.push = self.shell.push
        elif isinstance(shell, (TerminalInteractiveShell, ZMQInteractiveShell)):
            # if launching from an ipython or jupyter then adding a console is
            # not supported. Instead users should use the existing interactive
            # terminal for the same functionality.
            self.kernel_client = None
            self.kernel_manager = None
            self.shell = None
            self.push = lambda var: None

        else:
            raise ValueError(f"ipython shell not recognized; got {type(shell)}")
        # Add any user variables
        self.push(variables)

        self.enable_calltips = False

        # Set stylings
        self._update_theme()

        # TODO: Try to get console from jupyter to run without a shift click
        # self.execute_on_complete_input = True

    def _update_theme(self, event: ty.Any = None) -> None:
        """Update the napari GUI theme."""
        from qtextra.config.theme import THEMES

        # qtconsole unfortunately won't inherit the parent stylesheet
        # so it needs to be directly set
        # template and apply the primary stylesheet
        # (should probably be done by napari)
        # After napari 0.4.11, themes are evented models rather than
        # dicts.
        self.style_sheet = THEMES.get_theme_stylesheet()

        # After napari 0.4.6 the following syntax will be allowed
        # self.style_sheet = get_stylesheet(self.viewer.theme)

        # Set syntax styling and highlighting using theme
        self.syntax_style = THEMES.syntax_style
        bracket_color = QColor(*str_to_rgb(THEMES.get_rgb_color("highlight")))
        self._bracket_matcher.format.setBackground(bracket_color)

    def closeEvent(self, event):
        """Clean up the integrated console in napari."""
        if self.kernel_client is not None:
            self.kernel_client.stop_channels()
        if self.kernel_manager is not None and self.kernel_manager.has_kernel:
            self.kernel_manager.shutdown_kernel()

        # RichJupyterWidget doesn't clean these up
        self._completion_widget.deleteLater()
        self._call_tip_widget.deleteLater()
        self.deleteLater()
        event.accept()


class QtConsoleDialog(QtFramelessTool):
    """Console popup."""

    HIDE_WHEN_CLOSE = True

    def __init__(self, parent: ty.Optional[QWidget]) -> None:
        super().__init__(parent)
        self.setMinimumWidth(600)
        self.setMinimumHeight(400)
        self.setWindowFlags(self.windowFlags() | Qt.WindowStaysOnTopHint)  # type: ignore

    # noinspection PyAttributeOutsideInit
    def make_panel(self) -> QFormLayout:
        """Make panel."""
        _, header_layout = self._make_hide_handle()
        self._title_label.setText("IPython Console")

        self._console = QtConsole()

        layout = QVBoxLayout()
        layout.addLayout(header_layout)
        layout.addWidget(self._console)
        layout.setSpacing(2)
        layout.setContentsMargins(6, 6, 6, 6)
        return layout

    def push_variables(self, variables: ty.Dict[str, ty.Any]) -> None:
        """Push variables to console."""
        self._console.push(variables)



import pytest
from qtpy.QtGui import QIcon
from qtpy.QtWidgets import QMenu

from qtextra.widgets.qt_button_tool import QtToolButton


@pytest.fixture
def make_menu(qapp):
    """Make menu."""

    def _wrap():
        menu = QMenu(None)
        menu.addAction("Action 1")
        return menu

    return _wrap


@pytest.fixture
def setup_widget(qtbot):
    """Set up error report dialog."""

    def _wrap():
        widget = QtToolButton(None, "test")
        qtbot.addWidget(widget)
        return widget

    return _wrap


def test_qt_tool_button(setup_widget, make_menu, get_icon_path, qtbot):
    widget = setup_widget()
    menu = make_menu()

    # check label
    assert widget.text() == "test"

    # check menu
    widget.set_menu(menu, print)
    assert widget.menu() == menu

    # check icon
    icon = None
    widget.set_icon(icon)
    assert widget.icon().isNull() is True

    # will check if the path was not returned as None
    if get_icon_path:
        icon = QIcon(get_icon_path)
        widget.set_icon(icon)
        assert widget.icon().availableSizes()[0] == icon.availableSizes()[0]

    widget.set_size((10, 10))
    size = widget.size()
    assert size.width() <= 10 and size.height() <= 10


def test_qt_tool_button_wrong(setup_widget, make_menu, get_icon_path, qtbot):
    widget = setup_widget()
    menu = make_menu()

    with pytest.raises(ValueError) as __:
        widget.set_menu(menu, "not a callable")

    with pytest.raises(ValueError) as __:
        widget.set_menu("not a menu", print)



"""System summary widget."""

from __future__ import annotations

import os
from contextlib import suppress

import numba
import psutil
from koyo.utilities import human_readable_byte_size
from numba.cuda import CudaSupportError
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QGroupBox, QLabel, QVBoxLayout, QWidget

import qtextra.helpers as hp
from qtextra.widgets.qt_dialog import QtFramelessPopup

MEM_USAGE_ERROR = 4e9  # 4 Gb
MEM_USAGE_WARNING = 16e9  # 16 Gb
MEM_ERROR = 8e9
MEM_WARNING = 32e9

CPU_N_ERROR = 30
CPU_N_WARNING = 15

CPU_PERCENT_ERROR = 50
CPU_PERCENT_WARNING = 30


def style_if(widget, value: int, error_if: int, warn_if: int, less: bool = True) -> None:
    """Style widget based on value."""
    if less:
        if value < error_if:
            object_name = "error_label"
        elif value < warn_if:
            object_name = "warning_label"
        else:
            object_name = "success_label"
    else:
        if value > error_if:
            object_name = "error_label"
        elif value > warn_if:
            object_name = "warning_label"
        else:
            object_name = "success_label"
    hp.update_widget_style(widget, object_name)


class QtSystemSummaryWidget(QWidget):
    """System information."""

    _update_gpu: bool = True

    def __init__(self, parent: QWidget | None = None):
        QWidget.__init__(self, parent)

        # CPU summary
        self.cpu_group_box = QGroupBox("CPU Summary")
        self.cpu_group_box_layout = QVBoxLayout()
        self.cpu_group_box_layout.setSpacing(0)
        self.cpu_group_box_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.cpu_group_box.setLayout(self.cpu_group_box_layout)

        # CPU freq
        self.cpu_freq_stats_label = QLabel("", self)
        self.cpu_group_box_layout.addWidget(self.cpu_freq_stats_label)

        # Number of cores
        self.nb_cores_label = QLabel("", self)
        self.cpu_group_box_layout.addWidget(self.nb_cores_label)
        self.cpu_load_label0 = QLabel("", self)
        self.cpu_group_box_layout.addWidget(self.cpu_load_label0)
        self.cpu_load_label1 = QLabel("", self)
        self.cpu_group_box_layout.addWidget(self.cpu_load_label1)
        self.cpu_load_label2 = QLabel("", self)
        self.cpu_group_box_layout.addWidget(self.cpu_load_label2)

        # Memory summary
        self.memory_group_box = QGroupBox("Memory Summary")
        self.memory_group_box_layout = QVBoxLayout()
        self.memory_group_box_layout.setSpacing(0)
        self.memory_group_box_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.memory_group_box.setLayout(self.memory_group_box_layout)

        virtual = psutil.virtual_memory()
        available = virtual.available
        total = virtual.total

        try:
            mem = psutil.Process().memory_info().rss
        except Exception:
            mem = 0
        self.process_memory_label = QLabel(
            f"App Memory:\t {human_readable_byte_size(mem)}, ({round(100 * mem / total, 2)}%)", self
        )
        self.memory_group_box_layout.addWidget(self.process_memory_label)
        style_if(self.process_memory_label, mem, MEM_USAGE_ERROR, MEM_USAGE_WARNING, less=False)

        self.free_memory_label = QLabel(
            f"Free Memory:\t {human_readable_byte_size(available)}, ({round(100 * available / total, 2)}%)",
            self,
        )
        self.memory_group_box_layout.addWidget(self.free_memory_label)
        style_if(self.free_memory_label, available, MEM_ERROR, MEM_WARNING)

        self.total_memory_label = QLabel(
            f"Total Memory:\t {human_readable_byte_size(total)}",
            self,
        )
        self.memory_group_box_layout.addWidget(self.total_memory_label)
        style_if(self.total_memory_label, total, MEM_ERROR, MEM_WARNING)

        # GPU summary
        self.gpu_group_box = QGroupBox("GPU Summary")
        self.gpu_group_box_layout = QVBoxLayout()
        self.gpu_group_box_layout.setSpacing(0)
        self.gpu_group_box_layout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.gpu_group_box.setLayout(self.gpu_group_box_layout)

        try:
            cuda_gpu_name = numba.cuda.get_current_device().name.decode()
        except CudaSupportError:
            cuda_gpu_name = "N/A"

        self.cuda_gpu_label = QLabel(f"CUDA GPU: \t\t{cuda_gpu_name}", self)
        self.gpu_group_box_layout.addWidget(self.cuda_gpu_label)

        hp.set_object_name(
            self.cuda_gpu_label, object_name="success_label" if cuda_gpu_name != "N/A" else "error_label"
        )

        self.cudatoolkit_label = QLabel("", self)
        self.gpu_group_box_layout.addWidget(self.cudatoolkit_label)
        self.gpu_memory_free_label = QLabel("", self)
        self.gpu_group_box_layout.addWidget(self.gpu_memory_free_label)
        self.gpu_memory_total_label = QLabel("", self)
        self.gpu_group_box_layout.addWidget(self.gpu_memory_total_label)

        layout = QVBoxLayout(self)
        layout.setAlignment(Qt.AlignmentFlag.AlignCenter)
        layout.addWidget(self.cpu_group_box)
        layout.addWidget(self.memory_group_box)
        layout.addWidget(self.gpu_group_box)

        hp.make_periodic_timer(self, self.update_all, delay=5000, start=True)
        self.update_all()

    def update_all(self) -> None:
        """Update all stats."""
        self.update_cpu()
        self.update_mem()
        self.update_gpu()

    def update_cpu(self) -> None:
        """Update CPU stats."""
        with suppress(FileNotFoundError):
            cpu = round(psutil.cpu_freq().current, 2)
        cpu = f"{cpu} Mhz" if cpu else "N/A"
        self.cpu_freq_stats_label.setText(f"Current CPU frequency:     \t {cpu}")

        # Number of cores
        n_cpu = os.cpu_count() or 1
        self.nb_cores_label.setText(f"Number of CPU cores:       \t {n_cpu // 2}")
        style_if(self.nb_cores_label, n_cpu // 2, CPU_N_ERROR, CPU_N_WARNING)

        cpu_load_values = [(elem * 16) for elem in psutil.getloadavg()]
        cpu_1min = "100.0+" if cpu_load_values[0] >= 100.0 else round(cpu_load_values[0], 2)
        self.cpu_load_label0.setText(f"CPU load over last 1 min:  \t {cpu_1min}%")
        cpu_5min = "100.0+" if cpu_load_values[1] >= 100.0 else round(cpu_load_values[1], 2)
        self.cpu_load_label1.setText(f"CPU load over last 5 mins: \t {cpu_5min}%")
        cpu_15min = "100.0+" if cpu_load_values[2] >= 100.0 else round(cpu_load_values[2], 2)
        self.cpu_load_label2.setText(f"CPU load over last 15 mins:\t {cpu_15min}%")

        style_if(self.cpu_load_label0, cpu_load_values[0], CPU_PERCENT_ERROR, CPU_PERCENT_WARNING)
        style_if(self.cpu_load_label1, cpu_load_values[1], CPU_PERCENT_ERROR, CPU_PERCENT_WARNING)
        style_if(self.cpu_load_label2, cpu_load_values[2], CPU_PERCENT_ERROR, CPU_PERCENT_WARNING)

    def update_mem(self) -> None:
        """Update memory stats."""
        virtual = psutil.virtual_memory()
        available = virtual.available
        total = virtual.total

        try:
            mem = psutil.Process().memory_info().rss
        except Exception:
            mem = 0
        self.process_memory_label.setText(
            f"App Memory:\t {human_readable_byte_size(mem)}, ({round(100 * mem / total, 2)}%)"
        )
        style_if(self.process_memory_label, mem, MEM_USAGE_ERROR, MEM_USAGE_WARNING, less=False)

        self.free_memory_label.setText(
            f"Free Memory:\t {human_readable_byte_size(available)}, ({round(100 * available / total, 2)}%)"
        )
        style_if(self.free_memory_label, available, MEM_ERROR, MEM_WARNING)

        self.total_memory_label.setText(f"Total Memory:\t {human_readable_byte_size(total)}")
        self.memory_group_box_layout.addWidget(self.total_memory_label)
        style_if(self.total_memory_label, total, MEM_ERROR, MEM_WARNING)

    def update_gpu(self) -> None:
        """Update GPU stats."""
        # cuda_toolkit = numba.cuda.cudadrv.nvvm.is_available()
        # self.cudatoolkit_label.setText(f"CUDA Toolkit: \t\t{'present' if cuda_toolkit else 'absent'}")
        # if numba.cuda.cudadrv.nvvm.is_available():
        #     self.cudatoolkit_label.setStyleSheet("QLabel {color: green;}")
        # else:
        #     self.cudatoolkit_label.setStyleSheet("QLabel {color: red;}")

        if not self._update_gpu:
            return

        try:
            cuda_memory_free = numba.cuda.current_context().get_memory_info().free
            cuda_memory_total = numba.cuda.current_context().get_memory_info().total
        except CudaSupportError:
            cuda_memory_free = 0
            cuda_memory_total = 0
            self._update_gpu = False

        self.gpu_memory_free_label.setText(f"Free GPU Memory: \t{human_readable_byte_size(cuda_memory_free)}")
        self.gpu_memory_total_label.setText(f"Total GPU Memory: \t{human_readable_byte_size(cuda_memory_total)}")
        if cuda_memory_total == 0:
            hp.set_object_name(self.gpu_memory_total_label, self.gpu_memory_free_label, object_name="error_label")
        else:
            if numba.cuda.current_context().get_memory_info().total < 8000000000:
                hp.set_object_name(self.gpu_memory_free_label, object_name="warning_label")
            else:
                hp.set_object_name(self.gpu_memory_free_label, object_name="success_label")

            if cuda_memory_free / cuda_memory_total < 0.4:
                hp.set_object_name(self.gpu_memory_free_label, object_name="error_label")
            elif cuda_memory_free / cuda_memory_total < 0.8:
                hp.set_object_name(self.gpu_memory_free_label, object_name="warning_label")
            else:
                hp.set_object_name(self.gpu_memory_free_label, object_name="success_label")


class SystemSummaryPopup(QtFramelessPopup):
    """Show summary of the system."""

    def __init__(self, parent: QWidget | None = None):
        super().__init__(parent=parent)

    def make_panel(self) -> QVBoxLayout:
        """Create widget."""
        layout = QVBoxLayout()
        layout.setSpacing(0)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(QtSystemSummaryWidget(self))
        return layout


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys

        from qtextra.utils.dev import qframe

        app, frame, ha = qframe(False)
        frame.setMinimumSize(600, 600)

        wdg = QtSystemSummaryWidget(parent=frame)
        ha.addWidget(wdg)

        frame.show()
        sys.exit(app.exec_())

    _main()



"""Progress bar."""

from qtpy.QtCore import Property, QPoint, QRect, QSize, Qt, QVariantAnimation, Signal
from qtpy.QtGui import QFontMetrics, QPainter, QPen
from qtpy.QtWidgets import QWidget

from qtextra.config import THEMES

# TODO: if the step description is multi-line, it will not be displayed properly


class QtStepProgressBar(QWidget):
    """Progress bar with steps.

    https://stackoverflow.com/questions/63004722/how-to-create-a-labelled-qprogressbar-in-pyside
    """

    evt_steps_changed = Signal(list)
    evt_value_changed = Signal(int)

    # Attributes
    RADIUS = 10
    LINE_WIDTH = 5
    HORIZONTAL_PADDING = 5

    def __init__(self, parent=None):
        super().__init__(parent)

        self._labels = []
        self._value = 0

        self._percentage_width = 0
        self._animation = QVariantAnimation(startValue=0.0, endValue=1.0)
        self._animation.setDuration(500)
        self._animation.valueChanged.connect(self.update)

    def get_labels(self) -> list[str]:
        """Get labels."""
        return self._labels

    def set_labels(self, labels: list[str]) -> None:
        """Set labels."""
        self._labels = labels
        self.evt_steps_changed.emit(self._labels)

    labels = Property(list, fget=get_labels, fset=set_labels, notify=evt_steps_changed)

    def get_value(self) -> int:
        """Get value."""
        return self._value

    def set_value(self, value: int) -> None:
        """Set current value."""
        if 0 <= value < len(self.labels) + 1:
            self._value = value
            self.evt_value_changed.emit(value)
            self.update()
            if self.value < len(self.labels):
                self._animation.start()

    value = Property(int, fget=get_value, fset=set_value, notify=evt_value_changed)

    def sizeHint(self) -> QSize:
        """Return the size hint for the widget."""
        return QSize(320, 120)

    def paintEvent(self, event):
        default_line_color = THEMES.get_qt_color("primary")
        incomplete_color = THEMES.get_qt_color("secondary")
        progress_color = THEMES.get_qt_color("success")
        complete_color = THEMES.get_qt_color("success")
        canvas_color = THEMES.get_qt_color("canvas")
        text_color = THEMES.get_qt_color("text")

        painter = QPainter(self)

        painter.setRenderHints(QPainter.RenderHint.Antialiasing)

        painter.fillRect(self.rect(), canvas_color)

        busy_rect = QRect(0, 0, self.width(), self.LINE_WIDTH)
        busy_rect.adjust(self.HORIZONTAL_PADDING, 0, -self.HORIZONTAL_PADDING, 0)
        busy_rect.moveCenter(self.rect().center())

        painter.fillRect(busy_rect, default_line_color)

        number_of_steps = len(self.labels)
        if number_of_steps == 0:
            return

        step_width = busy_rect.width() / number_of_steps
        x = round(self.HORIZONTAL_PADDING + step_width / 2)
        y = round(busy_rect.center().y())

        r = QRect(0, 0, round(1.5 * self.RADIUS), round(1.5 * self.RADIUS))

        font_text = painter.font()
        fm = QFontMetrics(font_text)

        for i, text in enumerate(self.labels, 1):
            r.moveCenter(QPoint(x, y))

            if i <= self.value:
                w = step_width if i < self.value else self._animation.currentValue() * step_width
                r_busy = QRect(0, 0, round(w), round(self.LINE_WIDTH))
                r_busy.moveCenter(busy_rect.center())

                if i < number_of_steps:
                    r_busy.moveLeft(x)
                    painter.fillRect(r_busy, progress_color)

                pen = QPen(complete_color)
                pen.setWidth(3)
                painter.setPen(pen)
                painter.setBrush(complete_color)
                painter.drawEllipse(r)
                painter.setPen(canvas_color)
                painter.setPen(complete_color)

            else:
                is_active = (self.value + 1) == i
                pen = QPen(incomplete_color if is_active else default_line_color)
                pen.setWidth(3)
                painter.setPen(pen)
                painter.setBrush(canvas_color)
                painter.drawEllipse(r)
                painter.setPen(progress_color if is_active else text_color)

            rect = fm.boundingRect(text)
            rect.moveCenter(QPoint(int(x), round(y + 2 * self.RADIUS)))
            painter.setFont(font_text)
            painter.drawText(rect, Qt.AlignmentFlag.AlignCenter, text)
            x = int(x + step_width)


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtpy.QtWidgets import QPushButton

    from qtextra.utils.dev import qframe

    app, frame, ha = qframe(False)
    frame.setMinimumSize(600, 600)

    progressbar = QtStepProgressBar()
    progressbar.labels = [
        "Step One",
        "Step Two",
        "Step Three",
        "Step Four",
        "Step Five",
        "Complete",
    ]
    ha.addWidget(progressbar)

    button = QPushButton("Next Step")
    ha.addWidget(button)

    def on_clicked():
        progressbar.value = (progressbar.value + 1) % (len(progressbar.labels) + 1)

    button.clicked.connect(on_clicked)

    frame.show()
    sys.exit(app.exec_())



"""Info widgets."""

from __future__ import annotations

import weakref
from enum import Enum
from typing import Union

from qtpy.QtCore import (
    QEasingCurve,
    QEvent,
    QObject,
    QParallelAnimationGroup,
    QPoint,
    QPropertyAnimation,
    QSize,
    Qt,
    QTimer,
    Signal,
)
from qtpy.QtGui import QColor, QIcon, QPainter
from qtpy.QtWidgets import QFrame, QGraphicsOpacityEffect, QHBoxLayout, QVBoxLayout, QWidget

import qtextra.helpers as hp
from qtextra.config import is_dark
from qtextra.utils.wrap import TextWrap
from qtextra.widgets.qt_label_icon import QtSeverityLabel


class ToastPosition(Enum):
    """info toast position."""

    TOP = 0
    BOTTOM = 1
    TOP_LEFT = 2
    TOP_RIGHT = 3
    BOTTOM_LEFT = 4
    BOTTOM_RIGHT = 5
    NONE = 6


TOAST_POSITION_DICT = {
    "top": ToastPosition.TOP,
    "bottom": ToastPosition.BOTTOM,
    "top_left": ToastPosition.TOP_LEFT,
    "top_right": ToastPosition.TOP_RIGHT,
    "bottom_left": ToastPosition.BOTTOM_LEFT,
    "bottom_right": ToastPosition.BOTTOM_RIGHT,
    "none": ToastPosition.NONE,
}


class QtInfoToast(QFrame):
    """Information bar."""

    evt_closed = Signal()

    def __init__(
        self,
        icon: Union[QIcon, str],
        title: str,
        content: str,
        orientation: Qt.Orientation = Qt.Orientation.Vertical,
        is_closable: bool = True,
        duration: int = 3000,
        position=ToastPosition.TOP_RIGHT,
        parent: QWidget | None = None,
    ):
        """
        Parameters
        ----------
        icon: QtInfoToastIcon |  QIcon | str
            the icon of info toast

        title: str
            the title of info toast

        content: str
            the content of info toast

        orientation: Qt.Orientation
            the layout direction of info toast, use `Qt.Horizontal` for short content

        is_closable: bool
            whether to show the close button

        duration: int
            the time for info toast to display in milliseconds. If duration is less than zero,
            info toast will never disappear.

        parent: QWidget
            parent widget
        """
        super().__init__(parent=parent)
        self.title = title
        self.content = content
        self.orientation = orientation
        self.icon = icon
        self.duration = duration
        self.is_closable = is_closable
        self.position = position

        self.titleLabel = hp.make_label(self, object_name="titleLabel", bold=True)
        self.contentLabel = hp.make_label(self, object_name="contentLabel", wrap=True, enable_url=True)

        self.closeButton = hp.make_qta_btn(self, "cross", func=self.close)
        self.closeButton.set_normal()
        self.closeButton.setCursor(Qt.CursorShape.PointingHandCursor)
        self.closeButton.setVisible(self.is_closable)

        self.iconWidget = QtSeverityLabel(self)
        self.iconWidget.set_normal()
        self.iconWidget.severity = icon
        hp.set_properties(self, {"type": icon, "dark": is_dark()})

        self.hBoxLayout = QHBoxLayout(self)
        self.textLayout = QHBoxLayout() if self.orientation == Qt.Orientation.Horizontal else QVBoxLayout()
        self.widgetLayout = QHBoxLayout() if self.orientation == Qt.Orientation.Horizontal else QVBoxLayout()

        self.opacity_effect = QGraphicsOpacityEffect(self)
        self.opacity_animation = QPropertyAnimation(self.opacity_effect, b"opacity", self)
        self.opacity_effect.setOpacity(1)
        self.setGraphicsEffect(self.opacity_effect)

        self._light_bg_color = None
        self._dark_bg_color = None

        self.__initLayout()

    def __initLayout(self):
        self.hBoxLayout.setContentsMargins(6, 6, 6, 6)
        self.hBoxLayout.setSizeConstraint(QVBoxLayout.SizeConstraint.SetMinimumSize)
        self.textLayout.setSizeConstraint(QHBoxLayout.SizeConstraint.SetMinimumSize)
        self.textLayout.setAlignment(Qt.AlignmentFlag.AlignTop)
        self.textLayout.setContentsMargins(1, 8, 0, 8)

        self.hBoxLayout.setSpacing(0)
        self.textLayout.setSpacing(5)

        # add icon to layout
        self.hBoxLayout.addWidget(self.iconWidget, 0, Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)

        # add title to layout
        self.textLayout.addWidget(self.titleLabel, 1, Qt.AlignmentFlag.AlignTop)
        self.titleLabel.setVisible(bool(self.title))

        # add content label to layout
        if self.orientation == Qt.Orientation.Horizontal:
            self.textLayout.addSpacing(7)

        self.textLayout.addWidget(self.contentLabel, 1, Qt.AlignmentFlag.AlignTop)
        self.contentLabel.setVisible(bool(self.content))
        self.hBoxLayout.addLayout(self.textLayout, stretch=True)

        # add widget layout
        if self.orientation == Qt.Orientation.Horizontal:
            self.hBoxLayout.addLayout(self.widgetLayout)
            self.widgetLayout.setSpacing(10)
        else:
            self.textLayout.addLayout(self.widgetLayout)
            self.textLayout.addStretch(True)

        # add close button to layout
        self.hBoxLayout.addSpacing(12)
        self.hBoxLayout.addWidget(self.closeButton, 0, Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft)

        self._adjustText()

    def __fadeOut(self):
        """Fade out."""
        self.opacity_animation.setDuration(200)
        self.opacity_animation.setStartValue(1)
        self.opacity_animation.setEndValue(0)
        self.opacity_animation.finished.connect(self.close)
        self.opacity_animation.start()

    def _adjustText(self):
        w = 900 if not self.parent() else (self.parent().width() - 50)

        # adjust title
        chars = max(min(w / 10, 120), 30)
        self.titleLabel.setText(TextWrap.wrap(self.title, chars, False)[0])

        # adjust content
        chars = max(min(w / 9, 120), 30)
        self.contentLabel.setText(TextWrap.wrap(self.content, chars, False)[0])
        self.adjustSize()

    def addWidget(self, widget: QWidget, stretch=0):
        """Add widget to info toast."""
        self.widgetLayout.addSpacing(6)
        align = (
            Qt.AlignmentFlag.AlignTop if self.orientation == Qt.Orientation.Vertical else Qt.AlignmentFlag.AlignVCenter
        )
        self.widgetLayout.addWidget(widget, stretch, Qt.AlignmentFlag.AlignLeft | align)

    def setCustomBackgroundColor(self, light, dark):
        """Set the custom background color.

        Parameters
        ----------
        light, dark: str | Qt.GlobalColor | QColor
            background color in light/dark theme mode
        """
        self._light_bg_color = QColor(light)
        self._dark_bg_color = QColor(dark)
        self.update()

    def eventFilter(self, obj, e: QEvent):
        if obj is self.parent():
            if e.type() in [QEvent.Type.Resize, QEvent.Type.WindowStateChange]:
                self._adjustText()

        return super().eventFilter(obj, e)

    def closeEvent(self, e):
        self.evt_closed.emit()
        self.deleteLater()
        e.ignore()

    def showEvent(self, e):
        self._adjustText()
        super().showEvent(e)

        if self.duration >= 0:
            QTimer.singleShot(self.duration, self.__fadeOut)

        if self.position != ToastPosition.NONE:
            manager = QtInfoToastManager.make(self.position)
            manager.add(self)

        if self.parent():
            self.parent().installEventFilter(self)

    def paintEvent(self, e):
        super().paintEvent(e)
        if self._light_bg_color is None:
            return

        painter = QPainter(self)
        painter.setRenderHints(QPainter.RenderHint.Antialiasing)
        painter.setPen(Qt.PenStyle.NoPen)

        if is_dark():
            painter.setBrush(self._dark_bg_color)
        else:
            painter.setBrush(self._light_bg_color)

        rect = self.rect().adjusted(1, 1, -1, -1)
        painter.drawRoundedRect(rect, 6, 6)

    @classmethod
    def new(
        cls,
        icon,
        title,
        content,
        orientation: Qt.Orientation = Qt.Orientation.Vertical,
        is_closable=True,
        duration: int = 3000,
        position=ToastPosition.TOP_RIGHT,
        parent=None,
        min_width: int = 0,
    ):
        """Create new toast."""
        w = QtInfoToast(icon, title, content, orientation, is_closable, duration, position, parent)
        if min_width > 0:
            w.setMinimumWidth(min_width)
        w.show()
        return w

    @classmethod
    def info(
        cls,
        title: str,
        content: str,
        orientation: Qt.Orientation = Qt.Orientation.Vertical,
        is_closable: bool = True,
        duration: int = 1000,
        position: ToastPosition = ToastPosition.TOP_RIGHT,
        parent: QWidget | None = None,
    ):
        """Info toast."""
        return cls.new("info", title, content, orientation, is_closable, duration, position, parent)

    @classmethod
    def success(
        cls,
        title: str,
        content: str,
        orientation: Qt.Orientation = Qt.Orientation.Vertical,
        is_closable: bool = True,
        duration: int = 1000,
        position: ToastPosition = ToastPosition.TOP_RIGHT,
        parent: QWidget | None = None,
    ):
        """Success info toast."""
        return cls.new("success", title, content, orientation, is_closable, duration, position, parent)

    @classmethod
    def warning(
        cls,
        title: str,
        content: str,
        orientation: Qt.Orientation = Qt.Orientation.Vertical,
        is_closable: bool = True,
        duration: int = 1000,
        position: ToastPosition = ToastPosition.TOP_RIGHT,
        parent: QWidget | None = None,
    ):
        """Warning info toast."""
        return cls.new("warning", title, content, orientation, is_closable, duration, position, parent)

    @classmethod
    def error(
        cls,
        title: str,
        content: str,
        orientation: Qt.Orientation = Qt.Orientation.Vertical,
        is_closable: bool = True,
        duration: int = 1000,
        position: ToastPosition = ToastPosition.TOP_RIGHT,
        parent: QWidget | None = None,
    ):
        """Error info toast."""
        return cls.new("error", title, content, orientation, is_closable, duration, position, parent)

    @classmethod
    def critical(
        cls,
        title: str,
        content: str,
        orientation: Qt.Orientation = Qt.Orientation.Vertical,
        is_closable: bool = True,
        duration: int = 1000,
        position: ToastPosition = ToastPosition.TOP_RIGHT,
        parent: QWidget | None = None,
    ):
        """Critical info toast."""
        return cls.new("critical", title, content, orientation, is_closable, duration, position, parent)


class QtInfoToastManager(QObject):
    """info toast manager."""

    _instance = None
    managers = {}

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super().__new__(cls, *args, **kwargs)
            cls._instance.__initialized = False

        return cls._instance

    def __init__(self):
        if self.__initialized:
            return
        super().__init__()

        self.spacing = 16
        self.margin = 24
        self._toast = weakref.WeakKeyDictionary()
        self._animation_groups = weakref.WeakKeyDictionary()
        self._slide_animations = []
        self._drop_animations = []
        self.__initialized = True

    def add(self, toast: QtInfoToast):
        """Add info toast."""
        p = toast.parent()  # type:QWidget
        if not p:
            return

        if p not in self._toast:
            p.installEventFilter(self)
            self._toast[p] = []
            self._animation_groups[p] = QParallelAnimationGroup(self)

        if toast in self._toast[p]:
            return

        # add drop animation
        if self._toast[p]:
            remove_animation = QPropertyAnimation(toast, b"pos")
            remove_animation.setDuration(200)

            self._animation_groups[p].addAnimation(remove_animation)
            self._drop_animations.append(remove_animation)

            toast.setProperty("remove_animation", remove_animation)

        # add slide animation
        self._toast[p].append(toast)
        slide_animation = self._create_slide_animation(toast)
        self._slide_animations.append(slide_animation)

        toast.setProperty("slide_animation", slide_animation)
        toast.evt_closed.connect(lambda: self.remove(toast))
        slide_animation.start()

    def remove(self, toast: QtInfoToast):
        """Remove info toast."""
        p = toast.parent()
        if p not in self._toast:
            return

        if toast not in self._toast[p]:
            return

        self._toast[p].remove(toast)

        # remove drop animation
        remove_animation = toast.property("remove_animation")
        if remove_animation:
            self._animation_groups[p].removeAnimation(remove_animation)
            self._drop_animations.remove(remove_animation)

        # remove slider animation
        slide_animation = toast.property("slide_animation")
        if slide_animation:
            self._slide_animations.remove(slide_animation)

        # adjust the position of the remaining info toasts
        self._update_drop_animation(p)
        self._animation_groups[p].start()

    def _create_slide_animation(self, toast: QtInfoToast):
        slide_animation = QPropertyAnimation(toast, b"pos")
        slide_animation.setEasingCurve(QEasingCurve.Type.OutQuad)
        slide_animation.setDuration(200)

        slide_animation.setStartValue(self._get_slider_start_position(toast))
        slide_animation.setEndValue(self._get_position(toast))

        return slide_animation

    def _update_drop_animation(self, parent):
        for bar in self._toast[parent]:
            animation = bar.property("remove_animation")
            if not animation:
                continue

            animation.setStartValue(bar.pos())
            animation.setEndValue(self._get_position(bar))

    def _get_position(self, toast: QtInfoToast, size=None) -> QPoint:
        """Return the position of info toast."""
        raise NotImplementedError

    def _get_slider_start_position(self, toast: QtInfoToast) -> QPoint:
        """Return the start position of slide animation."""
        raise NotImplementedError

    def eventFilter(self, obj, e: QEvent):
        if obj not in self._toast:
            return False

        if e.type() in [QEvent.Type.Resize, QEvent.Type.WindowStateChange]:
            size = e.size() if e.type() == QEvent.Type.Resize else None
            for bar in self._toast[obj]:
                bar.move(self._get_position(bar, size))

        return super().eventFilter(obj, e)

    @classmethod
    def register(cls, name):
        """Register menu animation manager.

        Parameters
        ----------
        name: Any
            the name of manager, it should be unique
        """

        def wrapper(mgr):
            if name not in cls.managers:
                cls.managers[name] = mgr

            return mgr

        return wrapper

    @classmethod
    def make(cls, position: ToastPosition):
        """Mask info toast manager according to the display position."""
        if position not in cls.managers:
            raise ValueError(f"`{position}` is an invalid animation type.")

        return cls.managers[position]()


@QtInfoToastManager.register(ToastPosition.TOP)
class TopQtInfoToastManager(QtInfoToastManager):
    """Top position info toast manager."""

    def _get_position(self, toast: QtInfoToast, size=None):
        p = toast.parent()

        x = (toast.parent().width() - toast.width()) // 2
        y = self.margin
        index = self._toast[p].index(toast)
        for bar in self._toast[p][0:index]:
            y += bar.height() + self.spacing

        return QPoint(x, y)

    def _get_slider_start_position(self, toast: QtInfoToast):
        pos = self._get_position(toast)
        return QPoint(pos.x(), pos.y() - 16)


@QtInfoToastManager.register(ToastPosition.TOP_RIGHT)
class TopRightQtInfoToastManager(QtInfoToastManager):
    """Top right position info toast manager."""

    def _get_position(self, toast: QtInfoToast, size=None):
        p = toast.parent()
        size = size or p.size()

        x = size.width() - toast.width() - self.margin
        y = self.margin
        index = self._toast[p].index(toast)
        for bar in self._toast[p][0:index]:
            y += bar.height() + self.spacing

        return QPoint(x, y)

    def _get_slider_start_position(self, toast: QtInfoToast):
        return QPoint(toast.parent().width(), self._get_position(toast).y())


@QtInfoToastManager.register(ToastPosition.BOTTOM_RIGHT)
class BottomRightQtInfoToastManager(QtInfoToastManager):
    """Bottom right position info toast manager."""

    def _get_position(self, toast: QtInfoToast, size=None) -> QPoint:
        p = toast.parent()
        size = size or p.size()

        x = size.width() - toast.width() - self.margin
        y = size.height() - toast.height() - self.margin

        index = self._toast[p].index(toast)
        for bar in self._toast[p][0:index]:
            y -= bar.height() + self.spacing

        return QPoint(x, y)

    def _get_slider_start_position(self, toast: QtInfoToast):
        return QPoint(toast.parent().width(), self._get_position(toast).y())


@QtInfoToastManager.register(ToastPosition.TOP_LEFT)
class TopLeftQtInfoToastManager(QtInfoToastManager):
    """Top left position info toast manager."""

    def _get_position(self, toast: QtInfoToast, size=None) -> QPoint:
        p = toast.parent()

        y = self.margin
        index = self._toast[p].index(toast)

        for bar in self._toast[p][0:index]:
            y += bar.height() + self.spacing

        return QPoint(self.margin, y)

    def _get_slider_start_position(self, toast: QtInfoToast):
        return QPoint(-toast.width(), self._get_position(toast).y())


@QtInfoToastManager.register(ToastPosition.BOTTOM_LEFT)
class BottomLeftQtInfoToastManager(QtInfoToastManager):
    """Bottom left position info toast manager."""

    def _get_position(self, toast: QtInfoToast, size: QSize = None) -> QPoint:
        p = toast.parent()
        size = size or p.size()

        y = size.height() - toast.height() - self.margin
        index = self._toast[p].index(toast)

        for bar in self._toast[p][0:index]:
            y -= bar.height() + self.spacing

        return QPoint(self.margin, y)

    def _get_slider_start_position(self, toast: QtInfoToast):
        return QPoint(-toast.width(), self._get_position(toast).y())


@QtInfoToastManager.register(ToastPosition.BOTTOM)
class BottomQtInfoToastManager(QtInfoToastManager):
    """Bottom position info toast manager."""

    def _get_position(self, toast: QtInfoToast, size: QSize = None) -> QPoint:
        p = toast.parent()
        size = size or p.size()

        x = (size.width() - toast.width()) // 2
        y = size.height() - toast.height() - self.margin
        index = self._toast[p].index(toast)

        for bar in self._toast[p][0:index]:
            y -= bar.height() + self.spacing

        return QPoint(x, y)

    def _get_slider_start_position(self, toast: QtInfoToast):
        pos = self._get_position(toast)
        return QPoint(pos.x(), pos.y() + 16)


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys
        from random import choice

        from qtextra.config import THEMES
        from qtextra.utils.dev import qframe

        def _popup_notif() -> None:
            pop = [QtInfoToast.info, QtInfoToast.success, QtInfoToast.warning, QtInfoToast.error, QtInfoToast.critical]
            pop = choice(pop)
            pop = pop(
                title="Title",
                content="Here is a message.\nA couple of lines long.\nAnother line",
                parent=frame,
                position=choice(list(ToastPosition)),
                duration=3000,
                orientation=Qt.Orientation.Vertical,
            )

            THEMES.set_theme_stylesheet(pop)

        app, frame, ha = qframe(False, set_style=True)
        frame.setMinimumSize(600, 600)

        btn2 = hp.make_btn(frame, "Create random notification")
        btn2.clicked.connect(_popup_notif)
        ha.addWidget(btn2)
        ha.addStretch(1)

        frame.show()
        sys.exit(app.exec_())

    _main()



"""Vertical label."""

from qtpy import QtCore, QtGui
from qtpy import QtWidgets as QtW


class QtVerticalLabel(QtW.QLabel):
    """Rotated label."""

    def __init__(self, *args):
        QtW.QLabel.__init__(self, *args)

    def paintEvent(self, event) -> None:
        painter = QtGui.QPainter(self)
        painter.translate(0, self.height())
        painter.rotate(-90)
        # calculate the size of the font
        fm = QtGui.QFontMetrics(painter.font())
        xoffset = int(fm.boundingRect(self.text()).width() / 2)
        yoffset = int(fm.boundingRect(self.text()).height() / 2)
        x = int(self.width() / 2) + yoffset
        y = int(self.height() / 2) - xoffset
        # because we rotated the label, x affects the vertical placement, and y affects the horizontal
        painter.drawText(y, x, self.text())
        painter.end()

    def minimumSizeHint(self):
        size = super().minimumSizeHint()
        return QtCore.QSize(size.height(), size.width())

    def sizeHint(self):
        size = super().sizeHint()
        return QtCore.QSize(size.height(), size.width())


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.utils.dev import qframe

    def _test():
        print("clicked")

    app, frame, ha = qframe(False)
    frame.setMinimumSize(600, 600)
    ha.addWidget(QtVerticalLabel("Test 1 with more text"), stretch=True)

    frame.show()
    sys.exit(app.exec_())



"""Pill button."""

from __future__ import annotations

import typing as ty

from koyo.secret import get_short_hash
from qtpy.QtCore import QSize, Qt, Signal, Slot  # type: ignore[attr-defined]
from qtpy.QtGui import QMouseEvent
from qtpy.QtWidgets import QFrame, QHBoxLayout, QLabel, QSizePolicy, QWidget

import qtextra.helpers as hp
from qtextra.widgets.qt_button_icon import QtImagePushButton
from qtextra.widgets.qt_layout_scroll import QtScrollableHLayoutWidget

# FIXME: There is a bug that only occurs when:
#       1. Remove single widget. Dont check any of the existing widgets.
#       2. Try adding new widget - it will crash.


class QtLeftPillLabel(QLabel):
    """Left label."""

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)


class QtPillActionButton(QtImagePushButton):
    """Delete button."""

    def __init__(self, *args: ty.Any, **kwargs: ty.Any):
        super().__init__(*args, **kwargs)
        self._mode = self._icon = "delete"
        self.mode = self.icon = "delete"

    @property
    def icon(self) -> str:
        """Get icon."""
        return self._icon

    @icon.setter
    def icon(self, value: str):
        """Set icon."""
        self._icon = value
        self.set_qta(value)

    @property
    def mode(self) -> str:
        """Get mode."""
        return self._mode

    @mode.setter
    def mode(self, value: str) -> None:
        self._mode = value
        self.setProperty("mode", value)
        hp.polish_widget(self)


class QtTagButton(QFrame):
    """Two-sided pill button.

    The left side is used to show text of some kind and the right has delete button
    """

    evt_action = Signal(str)
    evt_clicked = Signal(str)
    evt_checked = Signal(str, bool)
    _active: bool = False

    def __init__(
        self,
        label: str,
        hash_id: str,
        parent: QWidget | None = None,
        allow_action: bool = True,
        action_type: str = "delete",
        action_icon: str = "cross",
        allow_selected: bool = True,
        hide_check: bool = False,
    ):
        super().__init__(parent=parent)
        self.setMaximumHeight(28)
        self.setMouseTracking(True)
        self.hash_id = hash_id
        self.hide_check = hide_check
        self._allow_selected = allow_selected
        self._label = label
        if hide_check:
            self.setProperty("hide_check", "True")

        self.selected = hp.make_qta_label(self, "check")
        self.selected.set_small()
        if not self._allow_selected:
            self.selected.evt_clicked.connect(self._handle_click)
        self.selected.setSizePolicy(QSizePolicy.Policy.MinimumExpanding, QSizePolicy.Policy.MinimumExpanding)

        self.label = QtLeftPillLabel(parent=self, text=label)
        self.label.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.MinimumExpanding)
        if not label:
            self.label.setVisible(False)

        self.action_btn = QtPillActionButton(parent=self)
        self.action_btn.set_xsmall()
        self.action_btn.clicked.connect(self._on_action)
        self.action_btn.setSizePolicy(QSizePolicy.Policy.Fixed, QSizePolicy.Policy.MinimumExpanding)
        self.action_btn.setVisible(allow_action)
        self.action_btn.icon = action_icon
        self.action_btn.mode = action_type
        self.setProperty("mode", action_type)
        hp.polish_widget(self.action_btn)

        layout = QHBoxLayout(self)
        layout.addWidget(self.selected, alignment=Qt.AlignmentFlag.AlignCenter)
        layout.addWidget(self.label, alignment=Qt.AlignmentFlag.AlignCenter, stretch=True)
        if allow_action:
            layout.addWidget(hp.make_v_line(self))
        layout.addWidget(self.action_btn, alignment=Qt.AlignmentFlag.AlignCenter)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.setSpacing(0)
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.MinimumExpanding)

        self.label.adjustSize()
        self.action_btn.adjustSize()
        self.adjustSize()

        self.active = self._active

    @property
    def text(self) -> str:
        """Get name of the tag."""
        return self.label.text()

    @text.setter
    def text(self, value: str) -> None:
        self.label.setText(value)
        self.label.setVisible(len(value) > 0)

    @property
    def active(self) -> bool:
        """Get checked state."""
        return self._active

    @active.setter
    def active(self, state: bool) -> None:
        self.setProperty("active", str(state))
        self.selected.setVisible(False if self.hide_check else state)
        hp.polish_widget(self)
        self._active = state
        self.evt_checked.emit(self.hash_id, state)

    def mousePressEvent(self, event: QMouseEvent) -> None:  # type: ignore[override]
        """Process mouse press event."""
        if event.button() == Qt.MouseButton.LeftButton:
            if self._allow_selected:
                self.active = not self._active
            else:
                self._handle_click()
        super().mousePressEvent(event)

    def _handle_click(self) -> None:
        self.evt_clicked.emit(self.hash_id)

    def sizeHint(self) -> QSize:
        """Get size hint."""
        sh = self.selected.sizeHint() + self.label.sizeHint()
        sh += self.action_btn.sizeHint() if self.action_btn.isVisible() else QSize(0, 0)
        return sh

    def _on_action(self) -> None:
        """On delete."""
        self.evt_action.emit(self.hash_id)


class QtTagManager(QWidget):
    """Manager class that contains multiple QtTagButtons."""

    evt_changed = Signal(str, bool)
    evt_checked = Signal(list)
    evt_clicked = Signal(str)
    evt_plus_clicked = Signal()

    # Widgets
    _plus_btn = None
    _clear_btn = None
    _filter_edit = None

    def __init__(
        self,
        parent: QWidget | None = None,
        allow_action: bool = False,
        flow: bool = True,
    ):
        super().__init__(parent=parent)
        self.allow_action = allow_action
        self.case_sensitive = False

        layout = hp.make_h_layout(parent=self, margin=0, spacing=0)
        self._layout = hp.make_flow_layout() if flow else QtScrollableHLayoutWidget()
        if flow:
            self._layout.setHorizontalSpacing(2)
            self._layout.setVerticalSpacing(2)
            layout.addLayout(self._layout)
        else:
            self._layout.setSpacing(2)
            layout.addWidget(self._layout)

        self._action_layout = hp.make_h_layout(margin=(2, 0, 0, 0), spacing=0)
        layout.addLayout(self._action_layout)

        self._layout.setSpacing(2)
        self._layout.setContentsMargins(2, 2, 2, 2)
        self.widgets: dict[str, QtTagButton] = {}
        self.buttons: dict[str, QWidget] = {}
        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)

    def add_tag(
        self,
        text: str,
        hash_id: str | None = None,
        allow_action: bool | None = None,
        active: bool = False,
        allow_check: bool = True,
        hide_check: bool = False,
    ) -> str:
        """Add tag to ."""
        if not hash_id:
            hash_id = get_short_hash()
        allow_action = self.allow_action if allow_action is None else allow_action
        widget = QtTagButton(
            text,
            hash_id,
            allow_action=allow_action,
            parent=self,
            allow_selected=allow_check,
            hide_check=hide_check,
        )
        widget.active = active
        widget.evt_action.connect(self.remove_tag)
        widget.evt_checked.connect(self._tag_changed)
        widget.evt_clicked.connect(self.evt_clicked.emit)

        self._layout.addWidget(widget)
        self.widgets[hash_id] = widget
        self._handle_filter_by()
        return hash_id

    def add_tags(
        self,
        options: list[str],
        allow_action: bool | None = None,
        allow_check: bool = True,
        hide_check: bool = False,
    ) -> None:
        """Add tags."""
        for tag in options:
            self.add_tag(tag, allow_action=allow_action, allow_check=allow_check, hide_check=hide_check)

    @Slot(str)  # type: ignore[misc]
    def remove_tag(self, hash_id: str) -> None:
        """Remove tag."""
        widget = self.widgets.pop(hash_id, None)
        if widget:
            hp.disconnect_event(widget, "evt_action", self.remove_tag)
            hp.disconnect_event(widget, "evt_checked", self._tag_changed)
            self._layout.removeWidget(widget)
            widget.deleteLater()

    def remove_tags(self) -> None:
        """Clear all options."""
        for widget in self.widgets.values():
            widget.deleteLater()
        self.widgets.clear()

    def update_label(self, hash_id: str, new_label: str) -> None:
        """Update label of specified tag."""
        tag = self.widgets[hash_id]
        tag.text = new_label

    def add_button(self, icon_name: str, tooltip: str = "") -> QtImagePushButton:
        """Add button."""
        button = hp.make_qta_btn(self, icon_name, tooltip=tooltip, small=True, standout=True)
        self._action_layout.addWidget(button)
        self.buttons[icon_name] = button
        return button

    def add_plus(self) -> QtImagePushButton:
        """Add plus button."""
        if self._plus_btn is not None:
            raise ValueError("Add button already exists.")

        self._plus_btn = self.add_button("add")
        self._plus_btn.clicked.connect(self._handle_add_click)
        return self._plus_btn

    def _handle_add_click(self) -> None:
        """Handle add click."""
        text = hp.get_text(self, "Type-in new label.", "New label")
        if text:
            self.add_tag(text, allow_action=self.allow_action)
            self.evt_plus_clicked.emit()

    def add_clear(self) -> QtImagePushButton:
        """Add plus button."""
        if self._clear_btn is not None:
            raise ValueError("Clear button already exists.")

        self._clear_btn = self.add_button("cross")
        self._clear_btn.clicked.connect(self.clear_selection)
        return self._clear_btn

    def add_filter(
        self, placeholder: str = "Type-in tag name...", max_width: int = 150, case_sensitive: bool = False
    ) -> None:
        """Add filter."""
        if self._filter_edit is not None:
            raise ValueError("Filter edit already exists.")

        self.case_sensitive = case_sensitive
        self._filter_edit = hp.make_line_edit(
            self, placeholder=placeholder, func_changed=self._handle_filter_by, func_clear=self._handle_filter_by
        )
        if max_width:
            self._filter_edit.setMaximumWidth(max_width)
        self._action_layout.addWidget(self._filter_edit)

    def _handle_filter_by(self) -> None:
        if self._filter_edit is None:
            return

        if self.case_sensitive:
            self._handle_case_sensitive_filter_by()
        else:
            self._handle_case_insensitive_filter_by()

    def _handle_case_sensitive_filter_by(self) -> None:
        text = self._filter_edit.text()
        for widget in self.widgets.values():
            widget.setHidden(text not in widget.text)

    def _handle_case_insensitive_filter_by(self) -> None:
        text = self._filter_edit.text().lower()
        for widget in self.widgets.values():
            widget.setHidden(text not in widget.text.lower())

    @Slot(str, bool)  # type: ignore[misc]
    def _tag_changed(self, hash_id: str, state: bool) -> None:
        """Tag was checked or unchecked."""
        self.evt_changed.emit(hash_id, state)
        self.evt_checked.emit(self.selected_options)

    @property
    def selected_ids(self) -> list[str]:
        """Get list of selected tags."""
        selected = []
        for hash_id, tag in self.widgets.items():
            if tag.active:
                selected.append(hash_id)
        return selected

    @property
    def selected_options(self) -> list[str]:
        """Get list of selected tags."""
        selected = []
        for _hash_id, tag in self.widgets.items():
            if tag.active:
                selected.append(tag.text)
        return selected

    def clear_selection(self) -> None:
        """Clear selections."""
        for widget in self.widgets.values():
            if widget.active:
                widget.active = False

    # Alias methods to offer Qt-like interface
    addTag = add_tag
    addTags = add_tags
    removeTag = remove_tag
    removeTags = remove_tags
    clearSelection = clear_selection
    updateLabel = update_label
    addButton = add_button
    addPlus = add_plus
    addClear = add_clear


if __name__ == "__main__":  # pragma: no cover

    def _main():  # type: ignore[no-untyped-def]
        import sys

        from qtextra.utils.dev import qframe

        app, frame, va = qframe(False)
        frame.setMinimumSize(400, 400)

        mgr = QtTagManager(allow_action=True)
        for i in range(5):
            mgr.add_tag(f"Tag number: {i}")
        mgr.add_filter()
        mgr.add_plus()
        mgr.add_tag("Tag number: 10", allow_check=False)
        va.addWidget(mgr, stretch=True)

        mgr = QtTagManager(allow_action=False)
        for i in range(5):
            mgr.add_tag(f"Tag number: {i}", hide_check=False)
        mgr.add_plus()
        mgr.add_tag("Tag number: 10", allow_check=False)
        va.addWidget(mgr, stretch=True)

        mgr = QtTagManager(allow_action=False, flow=False)
        for i in range(5):
            mgr.add_tag(f"Tag number: {i}")
        mgr.add_filter()
        mgr.add_plus()
        mgr.add_clear()
        mgr.add_tag("Tag number: 10", allow_check=False)
        va.addWidget(mgr, stretch=True)

        widget = QtTagButton("Tag 1", "TEST", frame)
        va.addWidget(widget)
        widget = QtTagButton("Much longer label", "TEST", frame)
        va.addWidget(widget, stretch=True)
        widget = QtTagButton("And this is even longer label", "TEST", frame)
        va.addWidget(widget, stretch=True)

        frame.show()
        sys.exit(app.exec_())

    _main()



from __future__ import annotations

from typing import TypeVar

from qtpy import QtWidgets as QtW
from qtpy.QtCore import Qt

_I = TypeVar("_I", bound=QtW.QGraphicsItem)


class QBaseGraphicsScene(QtW.QGraphicsScene):
    def __init__(self, parent=None):
        super().__init__(parent)
        self._grab_source: QtW.QGraphicsItem | None = None

    def grabSource(self) -> QtW.QGraphicsItem | None:
        return self._grab_source

    def setGrabSource(self, item: QtW.QGraphicsItem | None):
        self._grab_source = item


class QBaseGraphicsView(QtW.QGraphicsView):
    def __init__(self):
        scene = QBaseGraphicsScene()
        super().__init__(scene)
        self.setAlignment(Qt.AlignmentFlag.AlignVCenter | Qt.AlignmentFlag.AlignHCenter)
        self.setVerticalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setHorizontalScrollBarPolicy(Qt.ScrollBarPolicy.ScrollBarAlwaysOff)
        self.setMouseTracking(True)

    def addItem(self, item: _I) -> _I:
        self.scene().addItem(item)
        return item

    def scene(self) -> QBaseGraphicsScene:
        return super().scene()



"""Table configuration class."""

from __future__ import annotations

import typing as ty

from koyo.containers import MutableMapping

ColumnSizing = ty.Literal["stretch", "fixed", "contents"]
Alignment = ty.Literal["left", "center", "right"]


class Column(ty.TypedDict):
    """Column data."""

    name: str
    tag: str
    type: str
    show: bool
    width: int
    order: int
    hidden: bool
    tooltip: str
    sizing: ColumnSizing


class TableConfig(MutableMapping[int, Column]):
    """Table configuration object."""

    def __init__(self, text_alignment: Alignment | str = "center") -> None:
        super().__init__()
        self._dict: dict[int, Column] = {}
        self.last_index = -1
        self.color_columns: list[int] = []
        self.no_sort_columns: list[int] = []
        self.checkable_columns: list[int] = []
        self.html_columns: list[int] = []
        self.icon_columns: list[int] = []
        self.text_alignment: Alignment = text_alignment

    def __getitem__(self, tag: ty.Union[int, str]) -> ty.Any:
        """Get item id."""
        if isinstance(tag, int):
            val = self._dict[tag]
        else:
            val = self.find_col_id(tag)
        if val == -1:
            raise KeyError("Could not retrieve value")
        return val

    def __iter__(self):
        return iter(self._dict)

    def __getattr__(self, item: ty.Union[int, str]) -> ty.Any:
        # allow access to group members via dot notation
        try:
            return self.__getitem__(item)
        except KeyError:
            raise AttributeError from None

    @property
    def n_columns(self) -> int:
        """Return number of columns."""
        return len(self)

    @property
    def header(self) -> list[str]:
        """Return header."""
        return [v["name"] for v in self.values()]

    @property
    def hidden_columns(self) -> list[int]:
        """Returns list of hidden columns."""
        return [value["order"] for value in self.values() if value["hidden"]]

    def update_attribute(self, name: str, attr: str, value: ty.Any) -> None:
        """Update attribute value."""
        for name_, meta_ in self.items():
            if name == name_:
                meta_[attr] = value

    def add(
        self,
        name: str,
        tag: str,
        dtype: str = "str",
        width: int = 0,
        show: bool = True,
        hidden: bool = False,
        is_color: bool = False,
        no_sort: bool = False,
        tooltip: str = "",
        sizing: ColumnSizing | str = "stretch",
        checkable: bool = False,
    ) -> TableConfig:
        """Add an item to the configuration."""
        if dtype == "bool":
            sizing = "contents"
        self.last_index += 1
        if self.last_index == 0 and dtype == "bool":
            checkable = True

        self[self.last_index] = {
            "name": name,
            "tag": tag,
            "type": dtype,
            "show": show,
            "width": width,
            "order": self.last_index,
            "hidden": hidden,
            "tooltip": tooltip,
            "sizing": sizing,
        }
        if is_color:
            self.color_columns.append(self.last_index)
        if checkable and dtype == "bool":
            self.checkable_columns.append(self.last_index)
        if no_sort:
            self.no_sort_columns.append(self.last_index)
        if dtype == "icon":
            self.icon_columns.append(self.last_index)
        return self

    def get_column(self, tag: str) -> Column | None:
        """Get column by tag."""
        for _col_id, col_info in self.items():
            if col_info["tag"] == tag:
                return col_info
        return None

    def find_col_id(self, tag: str) -> int:
        """Find column id by the tag."""
        for col_id, col_info in self.items():
            if col_info["tag"] == tag:
                return col_id
        return -1

    def get_width(self, column_id: int) -> int:
        """Get the width of column."""
        data = self.get(column_id, {})
        width: int = data.get("width", 100)
        return width

    def to_columns(self, include_check: bool = True) -> list[str]:
        """Return columns."""
        if include_check:
            return [v["name"] for v in self.values()]
        return [v["name"] for v in self.values() if v["tag"] != "check"]

    def column_iter(self) -> ty.Iterator[tuple[int, Column]]:
        """Return column iterator."""
        for col_id in self:
            yield col_id, self[col_id]



"""Init."""

from qtextra.config.events import EVENTS
from qtextra.config.theme import CANVAS, THEMES, CanvasThemes, Themes, is_dark


def get_settings():
    """Get settings."""
    pass


__all__ = ["CANVAS", "EVENTS", "THEMES", "CanvasThemes", "Themes", "get_settings", "is_dark"]



"""Feedback dialog."""

import getpass
import os
import typing as ty

from loguru import logger
from qtpy.QtCore import Qt
from qtpy.QtWidgets import QFormLayout, QWidget

import qtextra.helpers as hp
from qtextra.widgets.qt_dialog import QtDialog

SENTRY_DSN: str = os.getenv("QTEXTRA_TELEMETRY_SENTRY_DSN", "")
ORGANIZATION_SLUG: str = os.getenv("QTEXTRA_TELEMETRY_ORGANIZATION", "")
PROJECT_SLUG: str = os.getenv("QTEXTRA_TELEMETRY_PROJECT", "")

FEEDBACK_URL = f"https://sentry.io/api/0/projects/{ORGANIZATION_SLUG}/{PROJECT_SLUG}/user-feedback/"


class FeedbackDialog(QtDialog):
    """Dialog to give the user an option to provide feedback."""

    def __init__(self, parent: ty.Optional[QWidget] = None):
        super().__init__(parent=parent, title="Feedback")
        self.setMinimumSize(600, 400)

    def accept(self):
        """Submit message."""
        self.on_apply()
        title = self.title.text()
        message = self.message.toPlainText()
        if not message:
            hp.warn_pretty(self, "Please write-in a message before continuing.")
            return
        event_id = submit_feedback(title, message)
        hp.toast(
            self.parent(),
            title="Feedback",
            message="Feedback submitted successfully."
            if event_id
            else "Feedback submission failed. Please try again later.",
            icon="success" if event_id else "error",
            position="top_left",
        )
        super().accept()

    # noinspection PyAttributeOutsideInit
    def make_panel(self) -> QFormLayout:
        """Dialog to provide feedback."""
        self.info = hp.make_label(
            self,
            "<b>Send feedback</b><br><br>We always welcome user feedback, whether its good or bad. Please type in"
            " your thoughts and send them onwards.<br>",
            alignment=Qt.AlignmentFlag.AlignHCenter,
        )
        self.name = hp.make_line_edit(self, getpass.getuser(), placeholder="Your name")

        self.email = hp.make_line_edit(self, "", placeholder="Your email address")

        self.title = hp.make_line_edit(self, "User feedback", placeholder="Feedback title")
        self.message = hp.make_text_edit(self, "", placeholder="Your feedback")

        self.submit_btn = hp.make_btn(self, "Submit", func=self.accept)
        self.cancel_btn = hp.make_btn(self, "Cancel", func=self.reject)

        btn_layout = hp.make_h_layout(self.submit_btn, self.cancel_btn)

        layout = hp.make_form_layout()
        layout.addRow(self.info)
        layout.addRow(hp.make_label(self, "Name (required)"), self.name)
        layout.addRow(hp.make_label(self, "Email address (optional)"), self.email)
        layout.addRow(hp.make_label(self, "Title"), self.title)
        layout.addRow(hp.make_label(self, "Message"))
        layout.addRow(self.message)
        layout.addRow(btn_layout)
        return layout


def submit_feedback(title: str, message: str, name: str = "", email: str = ""):
    """Submit feedback."""
    import getpass

    import requests
    import sentry_sdk

    with sentry_sdk.push_scope() as scope:
        scope.set_extra("name", name or getpass.getuser())
        scope.set_extra("email", email or "unknown@unknown.com")
        scope.set_extra("message", message)
    event_id = sentry_sdk.capture_message(message=title, level="debug", scope=scope)
    if event_id:
        data = {
            "comments": message,
            "event_id": event_id,
            "email": email or "unknown@unknown.com",
            "name": name or "",
        }

        res = requests.post(
            url=FEEDBACK_URL,
            data=data,
            headers={"Authorization": f"DSN {SENTRY_DSN}"},
            timeout=5,
        )
        logger.trace(f"Submitted extra feedback. Status code: {res.status_code}.")
    logger.debug(f"Submitted feedback. Return: {event_id}")
    return event_id


if __name__ == "__main__":  # pragma: no cover
    import sys

    from qtextra.dialogs.sentry import install_error_monitor
    from qtextra.utils.dev import apply_style, qapplication

    app = qapplication(1)
    install_error_monitor()
    dlg = FeedbackDialog(None)
    apply_style(dlg)
    dlg.show()
    sys.exit(app.exec_())



"""QProcess wrapper."""

from __future__ import annotations

import typing as ty
from contextlib import suppress
from queue import Empty, SimpleQueue
from time import time as time_

from koyo.system import IS_WIN
from loguru import logger as logger_
from qtpy.QtCore import QObject, QProcess, QTimer, Signal  # type: ignore[attr-defined]

from qtextra.queue.task import Task
from qtextra.queue.utilities import _safe_call, escape_ansi, iterable_callbacks
from qtextra.typing import Callback, TaskState


def run_command(command: list[str]) -> None:
    """Execute command using the QProcess wrapper."""
    from qtextra.helpers import get_main_window

    program, args = command[0], command[1:]
    logger_.trace(f"Running command: {program} {' '.join(args)}")

    process = QProcess(get_main_window())
    process.finished.connect(process.deleteLater)
    process.finished.connect(lambda exit_code, exit_status: logger_.trace(f"Command finished with {exit_code}"))
    process.setProgram(program)
    # Under Windows, the `setArguments` arguments are wrapped in a string which renders the arguments
    # incorrect. It's safer to simply join the arguments  together and set them as one long string. The
    # assumption is that the arguments were properly setup in the first place!
    if IS_WIN and hasattr(process, "setNativeArguments"):
        process.setNativeArguments(" ".join(args))
    else:
        process.setArguments(args)
    process.start()


def decode(text: bytes) -> str:
    """Decode text."""
    try:
        return text.decode("utf-8-sig")
    except UnicodeDecodeError:
        return text.decode("latin1")


class Queue(SimpleQueue):
    """Queue class with few missing methods."""

    def clear(self) -> None:
        """Clear queue."""
        while not self.empty():
            try:
                self.get_nowait()
            except Empty:
                break


class QueueCommand:
    """Command."""

    def __init__(self, task_id: str, command_index: int, command_args: list[str]) -> None:
        """Initialize."""
        self.task_id = task_id
        self.command_index = command_index
        self.command_args = command_args


class QueueTask:
    """Task."""

    def __init__(self, index: int, task: Task) -> None:
        """Initialize."""
        self.index = index
        self.task = task


class QProcessWrapper(QObject):
    """Wrapper around QProcess that handles multiple tasks."""

    # signals
    evt_started = Signal(Task)
    evt_next = Signal(Task)
    evt_finished = Signal(Task)
    evt_errored = Signal(Task)
    evt_part_errored = Signal(Task)
    evt_progress = Signal(Task)
    evt_paused = Signal(Task, bool)
    evt_cancelled = Signal(Task)

    # properties
    _paused: bool = False
    _cancelled: bool = False
    _master_started: bool = False
    _master_finished: bool = False
    _task_queue_populated: bool = False
    _cancel_emitted: bool = False
    n_running: int = 0

    def __init__(
        self,
        parent: QObject,
        task: Task,
        func_start: ty.Optional[Callback] = None,
        func_error: ty.Optional[Callback] = None,
        func_end: ty.Optional[Callback] = None,
        func_post: ty.Optional[Callback] = None,
    ):
        """Initialize."""
        super().__init__(parent=parent)
        self.logger = logger_.bind(src=task.task_id)
        self.task = task

        # setup process
        self.process = QProcess(parent=self)
        self.process.setProcessChannelMode(QProcess.ProcessChannelMode.MergedChannels)
        self.process.started.connect(self.on_started)
        self.process.finished.connect(self.on_finished)
        self.process.errorOccurred.connect(self.on_error)
        self.process.readyReadStandardOutput.connect(self.on_stdout)
        # self.process.readyReadStandardError.connect(self.on_stderr)

        self.finished_tasks: ty.Set[str] = set()
        self.command_queue: SimpleQueue = Queue()
        self.current_task_id: ty.Optional[str] = None

        # setup functions
        self.func_start = iterable_callbacks(func_start)
        self.func_error = iterable_callbacks(func_error)
        self.func_end = iterable_callbacks(func_end)
        self.func_post = iterable_callbacks(func_post)

    def on_stdout(self) -> None:
        """Record stdout emitted by the process."""
        with suppress(KeyError, RuntimeError):
            if self.task:
                text = self.process.readAllStandardOutput().data()
                try:
                    text = decode(text)  # type: ignore[assignment]
                    if text:
                        self.task.append_output(escape_ansi(text))  # type: ignore[arg-type]
                        self.evt_progress.emit(self.task)  # type: ignore[unused-ignore]
                except UnicodeDecodeError as e:
                    self.logger.trace(f"Could not decode stdout. {e} - {text}")  # type: ignore[str-bytes-safe]
                    return

    def on_stderr(self) -> None:
        """Record stderr emitted by the process."""
        with suppress(KeyError, RuntimeError):
            if self.task:
                text = self.process.readAllStandardError().data()
                text = decode(text)  # type: ignore[assignment]
                if text:
                    self.task.append_output(escape_ansi(text))  # type: ignore[arg-type]
                    self.evt_progress.emit(self.task)  # type: ignore[unused-ignore]

    @property
    def master_started(self) -> bool:
        """Return bool if a task had started."""
        return self._master_started

    @master_started.setter
    def master_started(self, value: bool) -> None:
        """Set task had started."""
        self._master_started = value
        self._master_finished = not value

    @property
    def master_finished(self) -> bool:
        """Return bool if a task had started."""
        return self._master_finished

    @property
    def task_id(self) -> str:
        """Return task id."""
        return self.task.task_id

    def is_running(self) -> bool:
        """Flag to indicate whether the task is running."""
        return bool(self.process.state() == QProcess.ProcessState.Running)

    def summary(self) -> str:
        """Return task summary."""
        return self.task.task_name

    def on_started(self) -> None:
        """The process has started."""
        self.task.state = TaskState.RUNNING
        _safe_call(self.func_start, self.task, "task_started")
        self.master_started = True
        self.evt_started.emit(self.task)  # type: ignore[unused-ignore]

    def populate_command_queue(self) -> ty.Optional[str]:
        """Add commands to queue."""
        if not self._task_queue_populated:
            self._task_queue_populated = True
            self.logger.trace(f"Populating CommandQueue for '{self.task.task_name}'...")
            # check whether the task was finished and locked
            if self.task.state == TaskState.FINISHED:
                self.logger.trace(f"Task '{self.task.task_name}' was locked and already finished. Moving on...")
                self.evt_next.emit(self.task)  # type: ignore[unused-ignore]
            # check whether the task was cancelled
            elif self.task.state == TaskState.CANCELLED:
                self.logger.trace(f"Task '{self.task.task_name}' was cancelled. Moving to the next task...")
                self.evt_next.emit(self.task)  # type: ignore[unused-ignore]
            # check whether task is running or queued
            elif self.task.state in [TaskState.RUNNING, TaskState.QUEUED]:
                # iterate over each command and execute it
                for index, command_args in enumerate(self.task.command_args()):
                    cmd = QueueCommand(self.task.task_id, index, command_args)
                    self.command_queue.put(cmd)
                self.logger.trace(f"Added {self.command_queue.qsize()} commands to CommandQueue.")
                # all sub-commands have been previously executed so can move to the next task.
                if self.command_queue.qsize() == 0:
                    self.task.state = TaskState.FINISHED
                    self.logger.trace(
                        f"Changed state of '{self.task.task_name}' to '{self.task.state.value}' (populate)"
                    )
                    return self.populate_command_queue()
                self.current_task_id = self.task.task_id
                return self.current_task_id
        else:
            self.logger.trace(f"CommandQueue for '{self.task.task_id}' was empty. Nothing else to do...")
        return None

    def on_setup_task(self) -> None:
        """Execute next task."""
        # check whether the queue is empty
        # if it's not, execute the next available command
        if not self.command_queue.empty():
            self.on_execute_task()
        # if it is, populate it with more commands from the next available task
        else:
            # populate the queue
            current_task_id = self.populate_command_queue()
            # if there were no more  tasks, the `current_task_id` will be None
            if current_task_id is None:
                self.logger.trace("There are no more tasks to execute. Finishing up...")
                self.master_started = False
                self.task.state = TaskState.FINISHED
                self.evt_next.emit(self.task)  # type: ignore[unused-ignore]
                self.evt_finished.emit(self.task)  # type: ignore[unused-ignore]
                self.logger.debug("All tasks finished.")
            # otherwise, a new task was retrieved and more commands were added to the queue
            else:
                # execute the next command
                self.on_execute_task()

    def on_execute_task(self) -> None:
        """Execute tasks in the queue."""
        try:
            cmd: QueueCommand = self.command_queue.get_nowait()
            command_args = cmd.command_args
            command_index = cmd.command_index
            command = " ".join(command_args)
            if not self.task:
                raise ValueError("Task not found.")
            # set start time
            if command_index == 0:
                self.task.start_time = time_()
            # activate master task
            # activate the current task
            if not self.task.is_active():
                self.task.activate()
            # update state so that it's running
            if self.task.state != TaskState.RUNNING:
                self.task.state = TaskState.RUNNING
                self.logger.trace(f"Changed state of '{self.task.task_name}' to '{self.task.state.value}' (execute)")

            # get program and commands
            program, args = command_args[0], command_args[1:]
            self.process.setProgram(program)
            # Under Windows, the `setArguments` arguments are wrapped in a string which renders the arguments
            # incorrect. It's safer to simply join the arguments  together and set them as one long string. The
            # assumption is that the arguments were properly setup in the first place!
            if IS_WIN:
                if hasattr(self.process, "setNativeArguments"):
                    self.process.setNativeArguments(" ".join(args))  # type: ignore[attr-defined]
                else:
                    self.logger.error("QProcess does not support setNativeArguments. Using setArguments instead.")
                    self.process.setArguments(args)
            else:
                self.process.setArguments(args)
            # update task info
            self.task.set_command_index(command_index)
            self.logger.trace(f"Executing: {self.task.task_name} / {command_index} : {command}")
            # start the next task
            self.evt_next.emit(self.task)  # type: ignore[unused-ignore]
            self.start()
        except Empty:
            self.logger.trace("The queue was empty. Going to try to get next task...")

    def on_finished(self, exit_code: int, exit_status: QProcess.ExitStatus) -> None:
        """The process has finished."""
        # handle non-zero exit code
        if exit_code != 0:
            self.logger.error(f"Task exited with non-zero exit code. ({exit_code}; {exit_status!s})")
            self.on_error(self.process.error())
            return

        # the task has finished
        task = self.task
        if task and self.master_started:
            # check if there are any other tasks in the queue
            if self.command_queue.empty():
                if task.state == TaskState.RUNNING:
                    task.state = TaskState.FINISHED
                    self.logger.trace(f"Changed state of '{task.task_name}' to '{task.state.value}' (finished)")
                task.lock()  # lock task
                self.finished_tasks.add(task.task_id)
                # self.evt_next.emit(self.task)  # type: ignore[unused-ignore]
            self.logger.trace(f"Task finished successfully: '{task.task_name}' (finished)")
            self.on_setup_task()

        # all tasks have finished
        if self.master_finished:
            if task:
                # update stats
                if task.state == TaskState.RUNNING:
                    task.state = TaskState.FINISHED
                    self.logger.trace(f"Changed state of '{task.task_name}' to '{task.state.value}' (all-finished)")
                task.end_time = time_()
                task.lock()  # lock task
                self.finished_tasks.add(task.task_id)
                self.logger.trace(f"Added '{task.task_name}' to finished tasks (all-finished).")
                self.logger.debug("Task finished successfully.")

    def on_error(self, _error: ty.Any) -> None:
        """Process has errored."""
        self.process.setProgram(None)  # type: ignore[arg-type]
        self.task.state = TaskState.FAILED
        if self.task.state != TaskState.FINISHED:
            self.task.state = TaskState.FAILED
            self.logger.trace(f"Changed state of '{self.task.task_name}' to '{self.task.state.value}' (error)")
            self.task.lock()  # lock task
        self.command_queue.clear()
        self.task.end_time = time_()
        self.master_started = False
        if self._cancelled:
            self.task.state = TaskState.CANCELLED
            self._emit_cancel()
        else:
            self.evt_errored.emit(self.task)  # type: ignore[unused-ignore]
        _safe_call(self.func_error, self.task, "task_errored")

    def run(self) -> None:
        """Execute task."""
        if self.master_started:
            self.logger.debug("Task has already started.")
            return
        self.master_started = True
        self.start()

    def start(self) -> None:
        """Start the process."""
        # if the task state has not been changed to started, let's don't do anything
        if not self.master_started:
            self.logger.debug("Task could not start as 'master_started' is not set.")
            return
        # this is the first process
        if not self.process.program():
            self.logger.debug("Setting-up task (start)...")
            self.on_setup_task()

        process_state = self.process.state()
        process_id = self.process.processId()
        if process_state == QProcess.ProcessState.NotRunning or process_id == 0:
            if self._paused:
                self.task.state = TaskState.PAUSED
                self.evt_paused.emit(self.task, self._paused)  # type: ignore[unused-ignore]
                self.logger.trace(f"Task '{self.task.task_name}' was successfully paused")
            elif self._cancelled:
                self.task.state = TaskState.CANCELLED
                self._emit_cancel()
            else:
                self.logger.trace(
                    f"Starting task. state={process_state!s}; process_id={process_id!s}; paused={self._paused!s};"
                    f" cancelled={self._cancelled!s}"
                )
                self.process.start()

    def pause(self, paused: bool) -> None:
        """Pause process."""
        self._paused = paused
        self.task.state = TaskState.PAUSING if paused else TaskState.RUNNING
        self.start()
        self.logger.debug(f"{'Pausing' if paused else 'Resuming'} '{self.task.task_name}'")

    def cancel(self) -> None:
        """Cancel process."""
        self._cancelled = True
        self.task.state = TaskState.CANCELLING if self.is_running() else TaskState.CANCELLED
        try:
            while self.command_queue.qsize() > 0:
                self.command_queue.get_nowait()
            self.logger.trace(f"Changed state of '{self.task.task_name}' to '{self.task.state.value}' (cancel)")
        except Empty:
            pass
        if self.is_running():
            kill_timer = QTimer(self.process)
            kill_timer.singleShot(3000, self.process.kill)  # wait 3 second for process to be killed
            self.process.terminate()
        else:
            self._emit_cancel()
        self.master_started = False
        self.logger.debug(f"Cancelling '{self.task.task_name}'")

    def _emit_cancel(self) -> None:
        if not self._cancel_emitted:
            self._cancel_emitted = True
            self.evt_cancelled.emit(self.task)



"""Widget with indicators."""

from __future__ import annotations

import typing as ty
from functools import partial

from loguru import logger
from qtpy.QtCore import Qt, Slot  # type: ignore[attr-defined]
from qtpy.QtWidgets import (  # type: ignore[attr-defined]
    QAction,
    QButtonGroup,
    QHBoxLayout,
    QSizePolicy,
    QStackedLayout,
    QStackedWidget,
    QToolBar,
    QVBoxLayout,
    QWidget,
)

import qtextra.helpers as hp
from qtextra.utils.utilities import connect
from qtextra.widgets.qt_button_icon import QtToolbarPushButton


class QtAboutWidget(QWidget):
    """About widget."""

    def __init__(self, title: str, description: str, docs_link: str | None = None, parent: QWidget | None = None):
        super().__init__(parent)

        self.title_label = hp.make_label(self, title, bold=True, wrap=True)
        self.description_label = hp.make_label(self, description, enable_url=True, wrap=True)
        self.docs_label = hp.make_label(self, docs_link if docs_link else "", enable_url=True, wrap=True)
        if docs_link is None:
            self.docs_label.setVisible(False)

        self._layout = QVBoxLayout(self)
        self._layout.addWidget(self.title_label)
        self._layout.addWidget(
            self.description_label, stretch=True, alignment=Qt.AlignmentFlag.AlignTop | Qt.AlignmentFlag.AlignLeft
        )
        self._layout.addWidget(self.docs_label)

        self.setSizePolicy(QSizePolicy.Policy.Expanding, QSizePolicy.Policy.Expanding)

    @classmethod
    def make_widget(cls, title: str, description: str, docs: str, parent: QWidget | None = None) -> QtAboutWidget:
        """Make widget."""
        return QtAboutWidget(title, description, docs, parent=parent)


# class QtAboutPopup(QtTransparentPopup):
#     """About popup."""
#
#     def __init__(self, text: str, parent: ty.Optional[QWidget] = None):
#         super().__init__(parent=parent)
#         self.setFocusPolicy(Qt.FocusPolicy.NoFocus)
#
#         self.label = hp.make_label(self, text)
#         layout = QVBoxLayout(self.frame)
#         layout.setContentsMargins(5, 5, 5, 5)
#         layout.addWidget(self.label)
#
#     def on_show(self, state: bool):
#         """Show popup."""
#         self.show() if state else self.hide()


class QtPanelWidget(QWidget):
    """Stacked panel widget."""

    def __init__(self, parent: QWidget | None = None, position: str = "left"):
        super().__init__(parent)

        self._about_stack = QWidget(self)
        self._about_stack.setMinimumWidth(0)
        self._about_stack.setMaximumWidth(0)
        self._about_layout = QStackedLayout()
        self._about_stack.setLayout(self._about_layout)
        self._about_stack.setVisible(False)

        self._stack = QStackedWidget(parent)
        self._stack.setContentsMargins(0, 0, 0, 0)

        self._buttons = QToolBar(self)
        self._buttons.setContentsMargins(0, 0, 0, 0)

        spacer = hp.make_spacer_widget()
        self._spacer = self._buttons.addWidget(spacer)

        self._group = QButtonGroup(self)
        self._button_dict: ty.Dict[QtToolbarPushButton, QAction] = {}
        self._hidden_dict: ty.Dict[QtToolbarPushButton, QAction] = {}

        # Widget setup
        self._buttons.setOrientation(Qt.Orientation.Vertical)
        self._group.setExclusive(True)
        self._group.buttonToggled.connect(self._toggle_widget)

        self._layout = QHBoxLayout()
        self._layout.setContentsMargins(0, 0, 0, 0)
        self._layout.setSpacing(0)
        self._layout.addWidget(self._buttons)
        if position == "left":
            self._layout.addWidget(self._about_stack)
        else:
            self._layout.addWidget(self._about_stack)

        self._about_stack.setContentsMargins(0, 0, 0, 0)

        self.setContentsMargins(0, 0, 0, 0)
        self.setLayout(self._layout)

    @property
    def stack_widget(self) -> QStackedWidget:
        """Get stack widget."""
        return self._stack

    def get_widget(self, name: str) -> QtToolbarPushButton | None:
        """Get widget."""
        for button in self._button_dict:
            if button.objectName() == name:
                return button
        return None

    def get_index(self, button: QtToolbarPushButton) -> int:
        """Get index."""
        for i, btn in enumerate(self._button_dict):
            if btn == button:
                return i
        return -1

    def add_widget(
        self,
        name: str,
        tooltip: str | None = None,
        widget: QWidget | None = None,
        location: str = "top",
        func: ty.Callable | None = None,
    ) -> QtToolbarPushButton:
        """Add widget to the stack.

        Parameters
        ----------
        name : str
            name of the object that will be used to select the icon
        tooltip : Optional[str]
            text that will be used to generate the tooltip information - it will be overwritten if the `widget`
            implements the `_make_html_description` method which auto-generates tooltip information in html-rich format
        widget : Optional[QWidget]
            widget that will be inserted into the stack
        location : str
            location of the button - allowed values include `top` and `bottom`. Typically, buttons that go to the
            `bottom` will be simple click-buttons without widgets associated with them.
        func : Optional[Callable]
            function that will be connected to the button click event
        """
        assert location in ["top", "bottom"], "Incorrect location provided - use `top` or `bottom`"
        if self.get_widget(name):
            logger.warning(f"Button with name '{name}' already exists")

        button: QtToolbarPushButton = hp.make_toolbar_btn(
            self,
            name,
            checkable=widget is not None,
            large=True,
            # icon_kwargs=dict(color_active=THEMES.get_hex_color("success")),
        )
        button.setObjectName(name)
        if tooltip:
            button.setToolTip(tooltip)

        # get action button
        self._button_dict[button] = self._add_before(button) if location == "top" else self._add_after(button)
        self._group.addButton(button)
        if widget:
            self.connect_widget(name, widget, tooltip)
        elif func:
            button.evt_click.connect(func)
        return button

    def connect_widget(self, name: str, widget: QWidget, tooltip: str | None = None) -> None:
        """Connect widget."""
        button = self.get_widget(name)
        if not button:
            logger.warning(f"Button with name '{name}' not found")
            return
        button.disconnect()
        button.setCheckable(True)
        index = self.get_index(button)

        # create custom tooltip if it's possible
        if hasattr(widget, "_make_html_description"):
            tooltip = widget._make_html_description()  # type: ignore[union-attr]

        # about_widget = None
        # if hasattr(widget, "_make_html_metadata"):
        #     about_widget = QtAboutWidget.make_widget(*widget._make_html_metadata(), parent=self._about_stack)
        # elif hasattr(widget, "_make_html_description"):
        #     tooltip = widget._make_html_description()
        if tooltip:
            button.setToolTip(tooltip)
        button.panel_widget = widget
        if hasattr(widget, "toggle_button"):
            widget.toggle_button = button
        if hasattr(widget, "evt_indicate"):
            connect(widget.evt_indicate, button.set_indicator)
        if hasattr(widget, "evt_indicate_about"):
            connect(widget.evt_indicate_about, button.set_indicator)
        self._stack.insertWidget(index, widget)
        if self._stack.count() == 1:
            self._toggle_widget(button, True)
        button.evt_click.connect(partial(self._toggle_widget, button, True))

    def _add_before(self, button: QtToolbarPushButton) -> QAction:
        """Add button after."""
        return self._buttons.insertWidget(self._spacer, button)  # type: ignore[return-value]

    def _add_after(self, button: QtToolbarPushButton) -> QAction:
        return self._buttons.addWidget(button)  # type: ignore[return-value]

    def _show_another(self, button: QtToolbarPushButton) -> None:
        """Show another widget if current button is disabled or hidden."""
        for btn in self._button_dict:
            if btn != button:
                btn.setChecked(True)
                break

    def _toggle_widget(self, button: QtToolbarPushButton, value: bool) -> None:
        """Toggle widget and show appropriate widget."""
        if button in self._hidden_dict:
            self._show_another(button)
            return

        for btn in self._button_dict:
            if btn != button:
                with hp.qt_signals_blocked(btn):
                    btn.setChecked(False)

        button.setChecked(value)
        button.repaint()
        button.set_indicator("")

        widget = button.panel_widget
        if value and widget:
            self._stack.setCurrentWidget(widget)
        if hasattr(button, "about_widget") and button.about_widget:
            self._about_layout.setCurrentWidget(button.about_widget)
        self._stack.setVisible(value)

        # This is a bit of a hack but it's required to force-update vispy canvas after changing to view the panel
        if value and widget and hasattr(widget, "update_after_activation"):
            hp.call_later(self, widget.update_after_activation, 50)

    def enable_widget(self, button: QtToolbarPushButton) -> None:
        """Enable widget."""
        if button in self._hidden_dict:
            action = self._hidden_dict.pop(button, None)
            if action:
                action.setVisible(True)

    def disable_widget(self, button: QtToolbarPushButton) -> None:
        """Disable widget."""
        if button in self._hidden_dict:
            logger.debug("Button is already hidden")
            return
        action = self._button_dict[button]
        self._hidden_dict[button] = action
        action.setVisible(False)

    def add_home_button(self) -> None:
        """Add home button."""
        button = self.add_widget("menu", "Show/hide information about the widgets.")
        button.evt_click.connect(self.show_about_stack)

    def show_about_stack(self, _: ty.Any) -> None:
        """Show about stack."""
        if self._about_stack.maximumWidth() == 0:
            start, end = 0, 250
        else:
            start, end = 250, 0
        hp.expand_animation(self._about_stack, start, end)


class QtPanelToolbar(QToolBar):
    """Toolbar."""

    def __init__(self, parent: QWidget | None = None, position: str = "left"):
        super().__init__(parent=parent)

        self._widget = QtPanelWidget(self, position=position)

        self.setWindowTitle("Toolbar")
        self.setMovable(False)
        self.setAllowedAreas(Qt.ToolBarArea.LeftToolBarArea | Qt.ToolBarArea.RightToolBarArea)
        self.setObjectName(position)
        self.addWidget(self._widget)
        self.setContentsMargins(0, 0, 0, 0)

    @property
    def stack_widget(self) -> QStackedWidget:
        """Get instance of the stack widget."""
        return self._widget._stack

    # noinspection PyMissingOrEmptyDocstring
    def add_widget(
        self,
        name: str,
        tooltip: str | None = None,
        widget: QWidget | None = None,
        location: str = "top",
        func: ty.Callable | None = None,
    ) -> QtToolbarPushButton:
        """Add widget to the toolbar."""
        return self._widget.add_widget(name, tooltip, widget, location=location, func=func)

    add_widget.__doc__ = QtPanelWidget.__doc__

    def connect_widget(
        self,
        name: str,
        widget: QWidget,
        tooltip: str | None = None,
    ) -> None:
        """Add widget to the toolbar."""
        self._widget.connect_widget(name, widget, tooltip)

    def set_disabled(self, button: QtToolbarPushButton, disable: bool) -> None:
        """Set widget as disabled."""
        if disable:
            self.disable_widget(button)
        else:
            self.enable_widget(button)

    def enable_widget(self, button: QtToolbarPushButton) -> None:
        """Enable widget."""
        self._widget.enable_widget(button)

    def disable_widget(self, button: QtToolbarPushButton) -> None:
        """Disable widget."""
        self._widget.disable_widget(button)

    @Slot()  # type: ignore[misc]
    def deactivate_all(self) -> None:
        """Deactivate all indicators."""
        for btn in self._widget._button_dict:
            btn.set_indicator("")
            btn.repaint()


if __name__ == "__main__":  # pragma: no cover fmt: off

    def _main():  # type: ignore[no-untyped-def]
        import sys
        from random import choice

        from qtextra.assets import QTA_MAPPING
        from qtextra.helpers import make_btn
        from qtextra.utils.dev import qmain, theme_toggle_btn
        from qtextra.widgets.qt_dialog import QtTab

        def _add_button() -> None:
            name = choice(list(QTA_MAPPING.keys()))
            indicator_type = choice(["warning", "", "success", "active"])
            pos = choice(["top", "bottom"])
            tooltip = "<p style='white-space:pre'><h1>This is a much longer line than the first</h1></p>"
            # """<p style''white-space:pre'><h2><b>MyList</b></h2></p>"""
            button = toolbar.add_widget(name, tooltip, QWidget() if pos == "top" else None, pos)
            button.set_indicator(indicator_type)

        def _add_widget() -> None:
            class Test(QtTab):
                _description: ty.ClassVar[dict] = {
                    "title": choice(
                        ["dimensionality reduction", "machine learning", "spatial", "spectral", "highlights"]
                    ),
                    "description": "ABOUT THE PANEL",
                }

                def make_panel(self):
                    """Panel."""
                    return QHBoxLayout()

            panel = Test(frame)
            name = choice(list(QTA_MAPPING.keys()))
            toolbar.add_widget(name, widget=panel)

        def _disable_btn():
            button = choice(list(toolbar._widget._button_dict.keys()))
            toolbar._widget.disable_widget(button)

        def _enable_btn():
            button = choice(list(toolbar._widget._hidden_dict.keys()))
            toolbar._widget.enable_widget(button)

        app, frame, ha = qmain(False)
        frame.setMinimumSize(600, 600)

        toolbar = QtPanelToolbar(frame)
        frame.addToolBar(Qt.ToolBarArea.LeftToolBarArea, toolbar)

        btn2 = make_btn(frame, "Click me to add widget")
        btn2.clicked.connect(_add_button)
        ha.addWidget(btn2)

        btn2 = make_btn(frame, "Click me to add panel")
        btn2.clicked.connect(_add_widget)
        ha.addWidget(btn2)

        ha.addWidget(theme_toggle_btn(frame))

        btn2 = make_btn(frame, "Enable button")
        btn2.clicked.connect(_enable_btn)
        ha.addWidget(btn2)

        btn2 = make_btn(frame, "Disable button")
        btn2.clicked.connect(_disable_btn)
        ha.addWidget(btn2)

        frame.show()
        sys.exit(app.exec_())

    _main()
